import numpy as np
from ...tools import functions_list_to_string
from ...DnaQuote import DnaQuote
from ..DnaSupplier import DnaSupplier


from Bio.Restriction import AllEnzymes

from ..builtin_constraints import (
    GcContentConstraint,
    NoPatternConstraint,
    SequenceLengthConstraint,
)
from ..builtin_pricings import PerBasepairPricing


class CommercialDnaOffer(DnaSupplier):
    """External/Commercial source of DNA"""

    class_description = "External DNA offer"
    operation_type = "order"
    report_fa_symbol = u""
    report_fa_symbol_plain = "shopping-cart"
    report_color = "#ffeeee"

    def __init__(
        self,
        name,
        pricing,
        sequence_constraints=(),
        lead_time=0,
        memoize=False,
    ):
        self.name = name
        self.sequence_constraints = sequence_constraints
        self.pricing = pricing
        self.lead_time = (
            lead_time if callable(lead_time) else (lambda *a: lead_time)
        )
        self.memoize = memoize
        self.memoize_dict = {}
        self.min_basepair_price = 0
        if hasattr(pricing, "min_basepair_price"):
            self.min_basepair_price = pricing.min_basepair_price

        if hasattr(pricing, "fixed_cost"):
            max_length = min(
                [
                    cst.max_length
                    for cst in sequence_constraints
                    if isinstance(cst, SequenceLengthConstraint)
                ],
                default=None,
            )
            if max_length is not None:
                self.min_basepair_price += pricing.fixed_cost / max_length

    def __repr__(self):
        return self.name

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False
    ):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        lead_time = self.lead_time(sequence)
        price = self.pricing(sequence)
        accepted = (max_lead_time == np.inf) or (lead_time <= max_lead_time)
        return DnaQuote(
            self, sequence, accepted=accepted, lead_time=lead_time, price=price
        )

    def get_best_lead_time_under_price_limit(self, sequence, max_price):

        lead_time = self.lead_time(sequence)
        price = self.pricing(sequence)
        return DnaQuote(
            self,
            sequence,
            price=price,
            lead_time=lead_time,
            accepted=price < max_price,
        )

    def additional_dict_description(self):
        constraints = functions_list_to_string(self.sequence_constraints)
        return {
            "price function": functions_list_to_string([self.pricing]),
            "sequence constraints": constraints,
        }

    @staticmethod
    def from_dict(data):
        constraints = []
        if "gc_range" in data:
            mini, maxi = data["gc_range"]
            if (mini, maxi) != (0, 100):
                constraints.append(
                    GcContentConstraint(0.01 * mini, 0.01 * maxi)
                )
        if "size_range" in data:
            mini, maxi = data["size_range"]
            constraints.append(SequenceLengthConstraint(mini, maxi))
        if "forbidden" in data:
            ENZYMES = list(str(e) for e in AllEnzymes)
            for pattern in data["forbidden"].split(", "):
                if pattern == "":
                    continue
                enzyme = None
                if pattern in ENZYMES:
                    enzyme = pattern
                    pattern = None
                constraints.append(
                    NoPatternConstraint(pattern=pattern, enzyme=enzyme)
                )

        return CommercialDnaOffer(
            name=data["name"],
            pricing=PerBasepairPricing(data["cost_per_bp"]),
            lead_time=data["lead_time"],
            memoize=data.get("memoize", False),
            sequence_constraints=constraints,
        )



from Bio import SeqIO
from ...DnaQuote import DnaQuote
from ..DnaSupplier import DnaSupplier



class PartsLibrary(DnaSupplier):
    """Class for collections of ready-to-assemble parts.


    This class is admitedly under-developed and could be expanded-subclassed
    to accomodate with the different kinds of registries etc.
    """

    class_description = "Parts Library"
    operation_type = "library"
    report_fa_symbol = u""
    report_fa_symbol_plain = "book"

    report_color = "#feeefe"
    collections_by_id = {}
    library_classes = {}

    def __init__(
        self,
        name,
        parts_dict=None,
        fasta_file=None,
        memoize=False,
        price_per_part=0,
        lead_time=0,
        sequence_constraints=(),
    ):
        self.name = name
        self.price_per_part = price_per_part
        self.lead_time = lead_time
        self.sequence_constraints = sequence_constraints
        if fasta_file is not None:
            parts_dict = {
                record.id: str(record.seq).upper()
                for record in SeqIO.parse(fasta_file, "fasta")
            }
        self.parts_dict = parts_dict
        self.inverted_parts_dict = {v: k for k, v in parts_dict.items()}
        self.sequences_set = set(self.inverted_parts_dict)
        self.memoize = memoize
        self.memoize_dict = {}

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False,
    ):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quotes

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
       """
        sequence = self.preprocess_sequence(sequence)
        if sequence in self.sequences_set:
            part_name = self.inverted_parts_dict[sequence]
            return DnaQuote(
                self,
                sequence,
                accepted=True,
                price=self.price_per_part,
                lead_time=self.lead_time,
                message="Part: " + part_name,
                metadata={"part_name": part_name},
            )

        return DnaQuote(
            self,
            sequence,
            accepted=False,
            message="Sequence not in the library",
        )

    def preprocess_sequence(self, sequence):
        """Can be used by subclasses e.g. to anonymize wildcard nucleotides"""
        return sequence

    def additional_dict_description(self):
        return {"flanks length": self.flanks_length}

    @classmethod
    def from_dict(cls, data):
        parameters = cls.collections_by_id[data["collection"]]
        library_class = parameters.pop("library_class")
        if library_class in cls.library_classes:
            library_class = cls.library_classes[library_class]

        def get(param, default):
            return data.get(param, parameters.get(param, default))

        return library_class(
            name=get("name", "library"),
            price_per_part=get("price_per_part", 0),
            lead_time=get("lead_time", 0),
            parts_dict=get("parts_dict", None),
            fasta_file=get("fasta_file", None),
            memoize=get("memoize", None),
        )


class GoldenGatePartsLibrary(PartsLibrary):
    """Library of parts for Golden Gate Assembly"""

    class_description = "Golden Gate parts library"

    def __init__(
        self,
        name,
        parts_dict=None,
        fasta_file=None,
        price_per_part=0,
        lead_time=0,
        flanks_length=7,
        memoize=False,
        sequence_constraints=(),
    ):
        PartsLibrary.__init__(
            self,
            name,
            parts_dict=parts_dict,
            fasta_file=fasta_file,
            memoize=memoize,
            sequence_constraints=sequence_constraints,
        )
        self.flanks_length = flanks_length

    def suggest_cuts(self, sequence):
        suggested_cuts = []
        # + 2 is because the cut falls in the middle of the 4bp linker:
        flank = self.flanks_length
        for part, part_sequence in self.parts_dict.items():
            segment = part_sequence[flank:-flank]
            i = sequence.find(segment)
            if i != -1:
                suggested_cuts += [i + 2, i + len(segment) - 2]
        return sorted(list(set(suggested_cuts)))

    def suggest_segments(self, sequence):
        suggested_segments = []
        # + 2 is because the cut falls in the middle of the 4bp linker:
        flank = self.flanks_length
        for part, part_sequence in self.parts_dict.items():
            segment = part_sequence[flank:-flank]
            i = sequence.find(segment)
            if i != -1:
                L = len(segment)
                suggested_segments.append(((i + 2, i + L - 2), part))
        return sorted(set(suggested_segments))

    @classmethod
    def preprocess_sequence(cls, sequence):
        """Can be used by subclasses e.g. to anonymize wildcard nucleotides"""
        return sequence[:6] + "N" + sequence[7:-7] + "N" + sequence[-6:]

    def additional_dict_description(self):
        return {
            "class": "Golden Gate parts library",
            "operation_type": "library",
            "flanks length": self.flanks_length,
        }


PartsLibrary.library_classes.update(
    {"library": PartsLibrary, "golden_gate": GoldenGatePartsLibrary}
)




"""DnaWeaver Reports implements reporting methods for Dna Weaver outputs."""

from .DnaSupplier import DnaSupplier
from .builtin_suppliers import (
    SequenceAdapter,
    CommercialDnaOffer,
    DnaAssemblyStation,
    DnaSuppliersComparator,
    PartsLibrary,
    GoldenGatePartsLibrary,
    PcrExtractionStation,
    PcrLinearizationStation,
)

DnaSupplier.default_suppliers_dict = {
    "commercial": CommercialDnaOffer,
    "assembly": DnaAssemblyStation,
    "library": PartsLibrary,
    "golden_gate_library": GoldenGatePartsLibrary,
    "pcr": PcrExtractionStation,
    "comparator": DnaSuppliersComparator,
    "main": DnaSuppliersComparator,
}

__all__ = [
    "DnaSupplier",
    "SequenceAdapter",
    "CommercialDnaOffer",
    "DnaAssemblyStation",
    "DnaSuppliersComparator",
    "PartsLibrary",
    "GoldenGatePartsLibrary",
    "PcrExtractionStation",
    "PcrLinearizationStation",
]



"""Pricings for suppliers such as CommercialDnaOffer"""


class PerBasepairPricing:
    """Class for pricings of the form X/basepair + fixed cost.

    Parameters
    ----------

    per_basepair_pricing
      Variable cost per basepair on top of the fixed cost

    fixed_cost
      Fixed cost added to the variable cost for each sequence.

    Examples
    --------

    >>> commercial_offer = CommercialDnaOffer(
    >>>     name="Company 1",
    >>>     pricing=PerBasepairPricing(0.14, fixed_cost=30)
    >>> )

    """

    def __init__(self, per_basepair_price, fixed_cost=0):
        self.per_basepair_price = per_basepair_price
        self.min_basepair_price = per_basepair_price
        self.fixed_cost = fixed_cost

    def __call__(self, sequence):
        return len(sequence) * self.per_basepair_price + self.fixed_cost

    def __str__(self):
        if self.fixed_cost != 0:
            return "$(%.02f + %.03f/bp)" % (
                self.fixed_cost,
                self.per_basepair_price,
            )
        return "$%.03f/bp" % self.per_basepair_price


class FixedCostPricing:
    """Class for pricings that return the same price for each sequence.

    Examples
    --------

    >>> commercial_offer = CommercialDnaOffer(
    >>>     name="Company 1",
    >>>     pricing=FixedCostPricing(5)
    >>> )

    """

    def __init__(self, fixed_cost):
        self.fixed_cost = fixed_cost

    def __call__(self, sequence):
        return self.fixed_cost

    def __str__(self):
        return "$%.03f/order" % self.fixed_cost



# -*- coding: utf-8 -*-

from collections import defaultdict

import numpy as np
from ..DnaQuote import DnaQuote
from ..biotools import SequenceString

from . import mixins


class DnaSupplier(
    mixins.JsonImportMixin, mixins.ConstraintsMixin, mixins.SupplyNetworkMixin
):
    """Base class for all DnaSuppliers, which are the elements of the supply
    networks used to define assembly problems in DnaWeaver."""

    min_basepair_price = 0

    def get_quote(
        self,
        sequence,
        max_lead_time=None,
        max_price=None,
        with_assembly_plan=False,
        time_resolution=1.0,
    ):
        """Return a DnaQuote with price, lead time, etc. for a given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        max_price (float)
          If provided, the quote returned is the least-lead-time quote
          whose price is below or equal to `max_price`.
          This is done using bisection and can be slow as it requires to
          re-compute the problem many times
          Note that either this parameter or `max_lead_time` must be None

        with_assembly_plan
          If True, the assembly plan is added to the quote

        time_resolution
          Time resolution for the bisecting search if `max_price` is not None.

        Returns
        -------

        A DnaQuote object.

        """
        if hasattr(sequence, 'seq'):
            # Use a simpler format which still remembers the topology
            sequence = SequenceString.from_record(sequence)

        if max_lead_time is None:
            max_lead_time = np.inf

        if self.memoize:
            args = (
                hash(sequence),
                max_lead_time,
                max_price,
                with_assembly_plan,
            )
            quote = self.memoize_dict.get(args, None)
            if quote is not None:
                return quote

        if not self.verify_constraints(sequence):
            quote = DnaQuote(
                self,
                sequence,
                accepted=False,
                message="Sequence does not pass constraints.",
            )

        elif max_price is not None:
            quote = self.get_best_lead_time_under_price_limit(
                sequence,
                max_price=max_price,
                time_resolution=time_resolution,
                with_assembly_plan=with_assembly_plan,
            )
        else:
            quote = self.get_best_price(
                sequence,
                max_lead_time=max_lead_time,
                with_assembly_plan=with_assembly_plan,
            )

        if self.memoize:
            self.memoize_dict[args] = quote

        return quote

    def __repr__(self):
        return self.name

    def get_best_lead_time_under_price_limit(
        self, sequence, max_price, time_resolution, with_assembly_plan=False,
    ):
        """Return the quote with fastest lead time under the budget constraint

        Parameters
        ----------



        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_price (float)
          If provided, the quote returned is the least-lead-time quote
          whose price is below or equal to `max_price`.
          This is done using bisection and can be slow as it requires to
          re-compute the problem many times
          Note that either this parameter or `max_lead_time` must be None

        with_assembly_plan
          If True, the assembly plan is added to the quote

        time_resolution
          Time resolution for the bisecting search if `max_price` is not None.

        """

        def f(max_lead_time):
            return self.get_quote(
                sequence,
                max_lead_time=max_lead_time,
                with_assembly_plan=with_assembly_plan,
            )

        quote = f(None)
        if (not quote.accepted) or (quote.price > max_price):
            return DnaQuote(
                self,
                sequence,
                accepted=False,
                message="Price over limit even without time limit.",
            )
        step_size = quote.lead_time / 2.0
        lead_time = quote.lead_time - step_size
        best_quote = quote
        while step_size > time_resolution:
            quote = f(lead_time)
            if quote.accepted and (quote.price <= max_price):
                best_quote = quote
                lead_time -= step_size
            else:
                lead_time += step_size
            step_size /= 2.0

        return best_quote

    def prepare_network_on_sequence(self, sequence):
        _edges, levels = self.compute_supply_graph()
        for level in levels:
            for source in level:
                if hasattr(source, "prepare_on_sequence"):
                    source.prepare_on_sequence(sequence)

    def dict_description(self):
        result = {
            "name": self.name,
            "operation_type": self.operation_type,
            "class": self.class_description,
            "_report_color": self.report_color,
            "_report_fa_symbol": self.report_fa_symbol,
            "_report_fa_symbol_plain": self.report_fa_symbol_plain,
        }
        result.update(self.additional_dict_description())
        return result

    def additional_dict_description(self):
        return {}



import matplotlib.pyplot as plt
from .AssemblyBlocksMixin import AssemblyBlocksMixin
from .AssemblyGraphMixin import AssemblyGraphMixin
from .ColorsMixin import ColorsMixin
from .TimelineMixin import TimelineMixin

from .plot_supply_graph import plot_supply_graph

class PlotsMixin(
    AssemblyBlocksMixin, AssemblyGraphMixin, ColorsMixin, TimelineMixin
):
    def write_figures_folder(self, figures_folder):
        def write_ax_as_pdf(ax, target):
            ax.figure.savefig(
                target.open("wb"), format="pdf", bbox_inches="tight"
            )
            plt.close(ax.figure)

        pos, ax = plot_supply_graph(self)
        write_ax_as_pdf(ax, figures_folder._file("supply_network.pdf"))
        plt.close(ax.figure)

        pos, ax = self.plot_assembly_graph(ax=None, textprops=None)
        write_ax_as_pdf(ax, figures_folder._file("assembly_graph.pdf"))

        _, ax_assembly_blocks = plt.subplots(1, figsize=(7, 4))
        assembly_blocks_ax, lg = self.plot_assembly_blocks(
            backend="matplotlib",
            plot_top_assembly=False,
            ax=ax_assembly_blocks,
            parts_offset=0.1,
            legend=True,
        )
        assembly_blocks_ax.figure.subplots_adjust(bottom=0.3)
        write_ax_as_pdf(
            assembly_blocks_ax, figures_folder._file("assembly_blocks.pdf")
        )




import json
from copy import deepcopy
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
from io import StringIO

from ..AssemblyPlanReport import AssemblyPlanReport


class ExportsMixin:
    def tree_as_list(self):
        """Return a list containing the current AssemblyOperation and all its
        sub-operations and their respective sub-operations.

        Said otherwise, it flattens the assembly tree into the list of all
        nodes.
        """
        result = [self]
        if self.assembly_plan is not None:
            result += sum(
                [
                    child.tree_as_list()
                    for segment, child in self.assembly_plan.items()
                ],
                [],
            )
        return result

    def assembly_plan_as_dict(self, as_json=False, json_indent=None):
        """Return a JSON-like version of the nested tree.

        Parameters
        ----------

        as_json
          If True, a JSON string is returned, else the result is a dict object.

        json_indent
          number of spaces in the JSON indentation (for pretty printing). The
          default None means that the JSON will be on one line (TODO: check).



        Returns
        -------

          {
          "id": self.id,
          "source": self.source.name,
          "price": self.price,
          "lead_time": self.lead_time,
          "sequence": self.sequence,
          "message": self.message,
          "metadata" = self.metadata,
          "assembly_plan": { (start1, end1): {(subquote_1)},
                             (start2, end2): {(subquote_2)},
                           }
         }
        """
        final_location = (
            self.final_location if hasattr(self, "final_location") else None
        )
        matching_segment = (
            self.matching_segment
            if hasattr(self, "matching_segment")
            else None
        )

        assembly_plan = []
        if self.assembly_plan is not None:
            for (segment, quote) in self.assembly_plan.items():
                quote_as_dict = quote.assembly_plan_as_dict()
                quote_as_dict["segment_start"] = segment[0]
                quote_as_dict["segment_end"] = segment[1]
                assembly_plan.append(quote_as_dict)
        tree = {
            "id": self.id,
            "source": self.source.name,
            "price": self.price,
            "lead_time": self.lead_time,
            "sequence": self.sequence,
            "message": self.message,
            "metadata": self.metadata,
            "assembly_plan": assembly_plan,
            "final_location": final_location,
            "matching_segment": matching_segment,
            "accepted": self.accepted,
        }

        if as_json:
            return json.dumps(tree, indent=json_indent)
        else:
            return tree

    def to_record(self, record=None, record_id=None):
        """Return a Biopython seqrecord of the quote.

        >>> record = to_record(solution)
        >>> # Let's plot with DnaVu:
        >>> from dnavu import create_record_plot
        >>> from bokeh.io import output_file, show
        >>> output_file("view.html")
        >>> plot = create_record_plot(record)
        >>> show(plot)
        """
        if record_id is None:
            record_id = self.id
        if record is None:
            record = SeqRecord(Seq(self.sequence, DNAAlphabet()), id=record_id)
        else:
            record = deepcopy(record)

        if self.assembly_plan is not None:
            features = [
                SeqFeature(
                    FeatureLocation(segment[0], segment[1], 1),
                    type="Feature",
                    qualifiers={
                        "name": quote.id,
                        "source": quote.source,
                        "price": quote.price,
                        "lead_time": quote.lead_time,
                    },
                )
                for segment, quote in self.assembly_plan.items()
            ]
            record.features = features + record.features
        return record

    def write_genbank(
        self, filename=None, filehandle=None, record=None, record_id=None
    ):
        record = self.to_record(record=record, record_id=record_id)
        if filename is not None:
            with open(filename, "w+") as f:
                SeqIO.write(record, f, "genbank")
        else:
            output = StringIO()
            SeqIO.write(record, output, "genbank")
            return output.getvalue()

    def to_assembly_plan_report(
        self, refine_fragments_locations=True, autocolor_quotes=True
    ):
        if refine_fragments_locations:
            self.compute_fragments_final_locations()
        if not self.full_assembly_plan_computed:
            self.compute_full_assembly_plan()
        report = AssemblyPlanReport(
            plan=self.assembly_plan_as_dict(),
            sources=self.source.dict_supply_graph(),
        )
        if autocolor_quotes:
            report.autocolor_quote_sources()
        return report



class ObjectDict(dict):
    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        dict.__getattribute__(self, key)

    def __setattr__(self, key, value):
        self[key] = value
        self.__dict__[key] = value

    @staticmethod
    def from_dict(d):
        obj = ObjectDict(
            {
                key: (
                    [
                        ObjectDict.from_dict(e) if isinstance(e, dict) else e
                        for e in value
                    ]
                    if isinstance(value, (list, tuple))
                    else (
                        ObjectDict.from_dict(value)
                        if isinstance(value, dict)
                        else value
                    )
                )
                for (key, value) in d.items()
            }
        )
        for key, value in obj.items():
            sanitized_key = key.replace(" ", "_").replace(".", "_")
            obj.__dict__[sanitized_key] = value
        return obj


from io import StringIO
from Bio.Seq import Seq
from Bio import SeqIO
import numpy as np


def complement(dna_sequence):
    """Return the complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"TACGGC"``.
    Uses BioPython for speed.
    """
    return str(Seq(dna_sequence).complement())


def reverse_complement(sequence):
    """Return the reverse-complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"GCCGTA"``.
    Uses BioPython for speed.
    """
    return complement(sequence)[::-1]


def random_dna_sequence(length, probas=None, seed=None):
    """Return a random DNA sequence ("ATGGCGT...") with the specified length.

    Parameters
    ----------

    length
      Length of the DNA sequence.

    proba
      Frequencies for the different nucleotides, for instance
      ``probas={"A":0.2, "T":0.3, "G":0.3, "C":0.2}``.
      If not specified, all nucleotides are equiprobable (p=0.25).

    seed
      The seed to feed to the random number generator. When a seed is provided
      the random results depend deterministically on the seed, thus enabling
      reproducibility

    """
    if seed is not None:
        np.random.seed(seed)
    if probas is None:
        sequence = np.random.choice(list("ATCG"), length)
    else:
        bases, probas = zip(*probas.items())
        sequence = np.random.choice(bases, length, p=probas)
    return "".join(sequence)


def load_record(filename, name="unnamed"):
    """Load a sequence file (genbank or fasta) as  a Biopython record."""
    if filename.lower().endswith(("gb", "gbk")):
        record = SeqIO.read(filename, "genbank")
    elif filename.lower().endswith(("fa", "fasta")):
        record = SeqIO.read(filename, "fasta")
    else:
        raise ValueError("Unknown format for file: %s" % filename)
    record.id = name
    record.name = name.replace(" ", "_")[:20]
    return record

def string_to_sequence(string):
    """Convert a string of a fasta, genbank... into a simple ATGC string.

    Can also be used to detect a format.

    """
    matches = re.match("([ATGC][ATGC]*)", string)
    if (matches is not None) and (matches.groups()[0] == string):
        return (string, "ATGC")

    for fmt in ("fasta", "genbank"):
        try:
            stringio = StringIO(string)
            return (str(SeqIO.read(stringio, fmt).seq), fmt)
        except Exception:
            pass
    raise ValueError("Invalid sequence format")

def file_to_sequence(filename):
    """Import a file in fasta, genbank... as a simple ATGC string."""
    with open(filename, "r") as f:
        return string_to_sequence(f.read())

def sequence_to_atgc(seq):
    if isinstance(seq, str):
        return seq
    if hasattr(seq, 'seq'):
        return str(seq.seq)
    return str(seq)


"""DnaWeaver, a Python package for optimal DNA sequence decomposition and
ordering."""

from .DnaSupplier import (
    DnaSupplier,
    CommercialDnaOffer,
    DnaSuppliersComparator,
    DnaAssemblyStation,
    PcrExtractionStation,
    PartsLibrary,
    SequenceAdapter,
    GoldenGatePartsLibrary,
    PcrLinearizationStation,
)
from .DnaAssemblyMethod import (
    GibsonAssemblyMethod,
    BuildAGenomeAssemblyMethod,
    GoldenGateAssemblyMethod,
    DnaAssemblyMethod
)
from .SegmentSelector import (
    SegmentSelector,
    FixedSizeSegmentSelector,
    TmSegmentSelector,
)
from .biotools import (
    random_dna_sequence,
    reverse_complement,
    string_to_sequence,
    load_record,
    gc_content,
    SequenceString,
    get_sequence_topology
)
from .DnaSupplier.builtin_constraints import (
    NoPatternConstraint,
    SequenceLengthConstraint,
    GcContentConstraint,
)
from .DnaSupplier.builtin_pricings import PerBasepairPricing, FixedCostPricing


import json
import networkx as nx

class JsonImportMixin:
    """Json import/export mixin for DnaSupplier"""

    @classmethod
    def from_json_data(
        cls, json_file=None, data=None, dna_suppliers_dict="default"
    ):
        """

        Returns
        -------

        levels, suppliers_dict, main_id
        """
        if dna_suppliers_dict == "default":
            dna_suppliers_dict = cls.default_suppliers_dict
        
        if json_file is not None:
            with open(json_file, 'r') as f:
                data = json.load(f)
        
        if "graph" in data:
            data = data["graph"]

        sorted_suppliers, levels = _sort_suppliers(data)
        main_id = sorted_suppliers[-1]
        suppliers_dict = {}
        for supplier_id in sorted_suppliers:
            supplier_data = data[supplier_id]
            supplier_data["parameters"]["name"] = supplier_data["name"]
            supplier_data["parameters"]["suppliers"] = [
                suppliers_dict[supp_id]
                for supp_id in supplier_data["suppliers"]
            ]
            supplier_class = dna_suppliers_dict[supplier_data["type"]]
            supplier = supplier_class.from_dict(supplier_data["parameters"])
            supplier.id = supplier_id
            suppliers_dict[supplier_id] = supplier
        main = suppliers_dict[main_id]
        main.data = {'levels': levels, 'suppliers_dict': suppliers_dict}
        return main

def _sort_suppliers(data):
    supply_graph = nx.DiGraph(
        [
            (supplier, supplier_id)
            for supplier_id, supplier_data in data.items()
            for supplier in supplier_data["suppliers"]
        ]
    )
    sorted_suppliers = []
    level = 0
    levels = {}
    while len(supply_graph):
        level += 1
        independant_suppliers = [
            n
            for n in supply_graph.nodes()
            if len(nx.ancestors(supply_graph, n)) == 0
        ]
        for supp in independant_suppliers:
            levels[supp] = level
        sorted_suppliers.extend(independant_suppliers)
        supply_graph.remove_nodes_from(independant_suppliers)
    return sorted_suppliers, levels





class DnaAssemblyMethod(object):
    """General class for assembly methods.

    All assembly methods can store the following attributes:

    Parameters
    ----------

    duration
      Duration required for the assembly (e.g. an estimated upper bound). The
      time unit is left at the choice of the user.

    cut_location_constraints
      List or tuple of functions `(sequence, int) -> bool` which return for
      a sequence and an index (cut location) whether the location should be
      considered as a cutting site compatible with this assembly method (True)
      or not.
      The locations considered in fine are the locations which pass every
      constraint in the `cut_location_contraints` list.

    segment_constraints
      List or tuple of functions `(sequence, start, end) -> bool` which returns
      for a subsegment `(start, end)` whether the segment is a valid segment
      for the assembly.
      The segments considered in fine are the segments which pass every
      constraint in the `segments_constraints` list.

    min_segment_length
      Minimal length of the fragments that this assembly method allows

    max_segment_length
      Maximal length of the fragments that this assembly method allows


    force_cuts
      A function `sequence->[i1, i2, i3...]` which for a given sequence returns
      forced cuts locations.

    max_fragments
       Maximal number of fragments which can be assembled with this method.

    sequence_constraints
      General constraints on the sequence so that it can be built with this
      assembly method. It is a list of functions `seq->bool`. If one of
      these constraints returns False the sequence is refused.

    reference
      Reference to e.g. a paper or a protocol describing the method.
    """

    name = "None"

    def __init__(
        self,
        duration=0,
        cost=0,
        reference=None,
        cut_location_constraints=(),
        segment_constraints=(),
        min_segment_length=0,
        max_segment_length=None,
        force_cuts=(),
        suggest_cuts=(),
        max_fragments=None,
        sequence_constraints=(),
        cuts_set_constraints=(),
    ):
        self.duration = duration
        self.cost = cost
        self.cut_location_constraints = list(cut_location_constraints)
        self.segment_constraints = list(segment_constraints)
        self.min_segment_length = min_segment_length
        self.max_segment_length = max_segment_length
        self.sequence_constraints = list(sequence_constraints)
        self.max_fragments = max_fragments
        self.cuts_set_constraints = list(cuts_set_constraints)
        self.reference = reference

        if callable(suggest_cuts):
            self.suggest_cuts = suggest_cuts
        else:
            # means the cuts are a constant
            self.suggest_cuts = lambda *a: suggest_cuts

        if callable(force_cuts):
            self.force_cuts = force_cuts
        else:
            # means the forced cuts are a constant
            self.force_cuts = lambda *a: force_cuts

    def dict_description(self):
        result = {
            "name": self.name,
            "minimum segment length": self.min_segment_length,
            "maximum segment length": self.max_segment_length,
            "maximal number of fragments": self.max_fragments,
            "reference": self.reference,
            "duration": self.duration,
            "cost": self.cost,
        }
        result.update(self.additional_dict_description())
        return result
    
    # def extend_sequence(self, sequence):
    #     return sequence

    def additional_dict_description(self):
        return {}



from .SegmentSelector import SegmentSelector


class FixedSizeSegmentSelector(SegmentSelector):
    """Selects segments of a constant size.

    Great for methods involving large homology regions where melting
    temperature matters less.
    """

    def __init__(self, segment_size=100, left_addition="", right_addition=""):
        self.segment_size = segment_size
        self.left_addition = left_addition
        self.right_addition = right_addition

    def compute_segment_location(self, sequence, index):
        return self.get_segment_coordinates(
            index, self.segment_size, len(sequence)
        )
    
    @property
    def max_homology_size(self):
        return self.segment_size

    def __str__(self):
        result = "FixedSize(%dbp)" % self.segment_size
        if self.left_addition:
            result = ("...%s-" % self.left_addition[-12:]) + result
        if self.right_addition:
            result = result + ("-%s..." % self.right_addition[:12])
        return result



from .DnaAssemblyMethod import DnaAssemblyMethod


class OverlapingAssemblyMethod(DnaAssemblyMethod):
    """General class for all overlaping assembly methods.

    Parameters
    ----------

    homology_arm_length
      Length of the homology arm, or "overhang". A length of L means that
      consecutive segments will overlap by 2*L

    """

    name = "Overlaping Assembly"

    def __init__(self, overhang_selector, **properties):
        super(OverlapingAssemblyMethod, self).__init__(**properties)
        selector = overhang_selector
        self.overhang_selector = selector
        if selector.has_location_filter:
            self.cut_location_constraints.append(
                selector.location_filter_method
            )
        self.compute_fragment_for_sequence_segment = (
            selector.compute_fragment_for_sequence_segment
        )

    # def extend_sequence(self, sequence):
    #     """Extend the end sequence with a start homology for circular assembly.
    #     """
    #     if self.topology == "circular":
    #         selector = self.overhang_selector
    #         start, end = selector.compute_segment_location(sequence, 0)
    #         size = end - start
    #         if size % 2:
    #             size += 1
    #         half_size = size // 2
    #         return sequence[-half_size:] + sequence + sequence[:half_size]
    #     else:
    #         return sequence


class GibsonAssemblyMethod(OverlapingAssemblyMethod):
    """Gibson Assembly Method. Just another overlap-method"""

    name = "Gibson Assembly"


class BuildAGenomeAssemblyMethod(OverlapingAssemblyMethod):
    """The Build-a-Genome Assembly Method. Just another overlap-method"""

    name = "Build-a-Genome"



"""Optimization techniques"""

import numpy as np
import networkx as nx
from .shortest_path_algorithms import (
    NoSolutionFoundError,
    shortest_valid_path,
    astar_path,
)
from proglog import default_bar_logger

# _default_bars = ('segment', 'edge')
# class SequenceDecomposerLogger(TqdmProgressBarLogger):

#     def __init__(self, bars=_default_bars, notebook='default',
#                  min_time_interval=0.2):
#         ignored_bars = set(_default_bars).difference(bars)
#         TqdmProgressBarLogger.__init__(self, bars=bars, notebook=notebook,
#                                        ignored_bars=ignored_bars,
#                                        min_time_interval=min_time_interval)


class SequenceDecomposer:
    """Find the sequence cuts which optimize the sum of segments scores.

    This is a very generic method meant to be applied to any sequence
    cutting optimization problem.

    Parameters
    ----------

    sequence_length
      Length of the sequence

    segment_score_function
      A function f( (start, end) ) -> score where (start, end) refers to the
      sequence segment start:end, and score is a float. The algorithm will
      produce cuts which minimize the total scores of the segments.

    cut_location_constraints
      List or tuple of functions `int -> bool` which return for each location
      whether the location should be considered as a cutting site (True) or
      discarded (False) in the decomposition of the sequence.
      The locations considered are the locations which pass every filter
      in the `cut_location_constraints` list.

    segment_constraints
      List or tuple of functions `(int, int) -> bool` which return for each
      segment `(start, end)` whether the segment is a valid segment for the
      decomposition of the sequence or whether it should be forbidden.
      The segmentss considered are the segments which pass every filter
      in the `segments_filters` list.

    forced_cuts
      List of locations at which the decomposition must imperatively cut, even
      if these cuts do not comply with the `cut_location_constraints`.

    max_segment_length
      Maximal length of the segments. Even though this could be specified using
      a filter in `segments_filters`, in practice providing this parameter
      accelerates computations as it allows to reduce the number
      of segments considered.

    a_star_factor
      If 0, the classical Dijkstra algorithm is used for path finding. Else,
      the a_star algorithm is used with a heuristic `h(x)=a_start_factor*(L-x)`
      where x is the location of a cutting point and L the length of the
      sequence. See the original DNAWeaver article for clearer explanations.
      Using a high A* factor can improve computing times several folds but
      yields suboptimal decompositions.

    path_size_limit
      Maximal number of edges for acceptable paths. None means no limit.
      Only works with `a_star_factor=0`

    path_size_min_step
      Minimal step for the bisection search when `path_size_limit` is not None
      (see `dijkstra_path_with_size_limit`)


    Returns
    -------

    graph
      The graph of the problem (a Networkx DiGraph whose nodes are indices of
      locations in the sequence and the "weight" of edge [i][j] is given by
      segment_score_function(segment[i][j])).

    best_cuts
      The list of optimal cuts (includes 0 and len(sequence)), which minimizes
      the total score of all the segments corresponding to the cuts.

    """

    def __init__(
        self,
        sequence_length,
        segment_score_function,
        cut_location_constraints=(),
        segment_constraints=(),
        forced_cuts=(),
        suggested_cuts=(),
        cuts_set_constraints=(),
        cut_spread_radius=0,
        min_segment_length=0,
        max_segment_length=None,
        coarse_grain=1,
        fine_grain=1,
        a_star_factor=0,
        path_size_limit=None,
        path_size_min_step=0.001,
        logger=None,
        bar_prefix="",
    ):

        self.segment_score_function = segment_score_function
        self.forced_cuts = set(list(forced_cuts) + [0, sequence_length])
        self.suggested_cuts = set(suggested_cuts)
        self.cut_spread_radius = cut_spread_radius
        self.segments_constraints = list(segment_constraints)
        self.cuts_set_constraints = list(cuts_set_constraints)
        self.cut_location_constraints = list(cut_location_constraints)
        self.sequence_length = sequence_length
        if max_segment_length is None:
            self.max_segment_length = sequence_length
        else:
            self.max_segment_length = max_segment_length
        self.min_segment_length = min_segment_length

        self.coarse_grain = coarse_grain
        self.fine_grain = fine_grain
        self.a_star_factor = a_star_factor
        self.path_size_limit = path_size_limit
        self.path_size_min_step = path_size_min_step
        self.logger = default_bar_logger(logger, min_time_interval=0.2)
        self.bar_prefix = bar_prefix

        if len(forced_cuts) > 0:

            def forced_cuts_filter(segment):
                start, end = segment
                return not any((start < cut < end) for cut in forced_cuts)

            self.segments_constraints.append(forced_cuts_filter)

        if len(self.cut_location_constraints) == 0:
            self.valid_cuts = set(range(sequence_length + 1))
        else:
            self.valid_cuts = set(
                [
                    index
                    for index in range(sequence_length + 1)
                    if all(fl(index) for fl in self.cut_location_constraints)
                ]
            )

    def compute_graph(
        self, valid_cuts, prune_deadends=True, reachable_indices_only=True
    ):
        L = self.sequence_length
        segments = []
        reachable_indices = set({0})

        # LIST THE SEGMENTS
        valid_cuts = sorted(valid_cuts)
        valid_cuts_array = np.zeros(L + 1)
        valid_cuts_array[valid_cuts] = 1
        for start in self.logger.iter_bar(
            segment=valid_cuts, bar_prefix=self.bar_prefix
        ):
            if reachable_indices_only and (start not in reachable_indices):
                continue
            ends_min = start + self.min_segment_length
            ends_max = min(L + 1, start + self.max_segment_length)
            hits = valid_cuts_array[ends_min:ends_max].nonzero()[0]
            valid_ends = [int(e) for e in hits + ends_min]
            if len(self.segments_constraints) > 0:
                valid_ends = [
                    end
                    for end in valid_ends
                    if all(
                        fl((start, end)) for fl in self.segments_constraints
                    )
                ]
            segments += [(start, end) for end in valid_ends]
            reachable_indices = reachable_indices.union(valid_ends)
        graph = nx.DiGraph(segments)

        if prune_deadends:
            ancestors = nx.ancestors(graph, L)
            graph.remove_nodes_from(
                [n for n in graph if (n != L) and (n not in ancestors)]
            )
        return graph

    def find_shortest_path(self, graph):
        constraints = self.cuts_set_constraints
        if (len(constraints) == 0) and (self.a_star_factor > 0):
            memodict = {}

            def compute_weight(start, end, props):
                """Compute the weight (cost) for segment (start, end).

                Parameter `props` is useless and is there for compatibility
                reasons
                """
                segment = tuple(sorted((start, end)))
                if segment in memodict:
                    return memodict[segment]
                score = self.segment_score_function(segment)
                if score >= 0:
                    result = score
                else:
                    result = np.inf
                memodict[segment] = result
                return result

            def heuristic(n1, n2):
                return self.a_star_factor * (n2 - n1)

            try:
                path = astar_path(
                    graph,
                    0,
                    self.sequence_length,
                    heuristic=heuristic,
                    weight=compute_weight,
                )
                return graph, path
            except (KeyError, nx.NetworkXNoPath):
                return graph, None

        else:
            for start, end in self.logger.iter_bar(
                edge=list(graph.edges()), bar_prefix=self.bar_prefix
            ):
                weight = self.segment_score_function((start, end))
                if weight < 0:
                    graph.remove_edge(start, end)
                else:
                    graph[start][end]["weight"] = weight

            try:
                path = shortest_valid_path(
                    graph,
                    0,
                    self.sequence_length,
                    nodes_constraints=constraints,
                    size_limit=self.path_size_limit,
                    min_step=self.path_size_min_step,
                )
                return graph, path
            except (KeyError, nx.NetworkXNoPath):
                return graph, None

    def compute_coarse_cuts(self, grain="default"):
        L = self.sequence_length
        if grain == "default":
            grain = self.coarse_grain
        grained_cuts = set(range(0, L + 1, grain))
        forced_but_invalid = self.forced_cuts.difference(self.valid_cuts)
        if len(forced_but_invalid):
            raise NoSolutionFoundError(
                "One forced cut was also invalid at indices %s"
                % forced_but_invalid
            )
        grained_and_suggested = grained_cuts.union(self.suggested_cuts)
        if self.cut_spread_radius:
            grained_and_suggested = spread_indices_in_set(
                indices=grained_and_suggested,
                radius=self.cut_spread_radius,
                max_index=L + 1,
            )
        valid_cuts = (
            grained_and_suggested
            .intersection(self.valid_cuts)
            .union(self.forced_cuts)
        )

        self.coarse_graph = self.compute_graph(valid_cuts)
        error = NoSolutionFoundError(
            "No coarse solution found, possibly too "
            "strong cuts/segments constraints."
        )
        if 0 not in self.coarse_graph.nodes():
            raise error
        graph, cuts = self.find_shortest_path(self.coarse_graph)

        self.weighted_coarse_graph = graph
        if cuts is None:
            raise error
        self.coarse_cuts = cuts

    def compute_fine_cuts(self):
        L = self.sequence_length
        radius = int(self.coarse_grain / 2)
        fine_cuts = set().union(
            *[
                set(
                    [cut]
                    + (
                        []
                        if cut in self.forced_cuts
                        else list(
                            range(
                                max(0, cut - radius),
                                min(L + 1, cut + radius),
                                self.fine_grain,
                            )
                        )
                    )
                )
                for cut in self.coarse_cuts
            ]
        )
        if self.cut_spread_radius:
            fine_cuts = spread_indices_in_set(
                indices=fine_cuts,
                radius=self.cut_spread_radius,
                max_index=L + 1,
            )
        fine_cuts = (fine_cuts.intersection(self.valid_cuts)).union(
            self.forced_cuts
        )
        self.fine_graph = self.compute_graph(fine_cuts)
        _, self.fine_cuts = self.find_shortest_path(self.fine_graph)

    def compute_optimal_cuts(self):
        self.compute_coarse_cuts()
        if (self.coarse_grain > 1) and self.fine_grain:
            self.compute_fine_cuts()
            return self.fine_cuts
        else:
            return self.coarse_cuts


def spread_indices_in_set(indices, radius, max_index):
    new_list = [
        new_index
        for i in indices
        for new_index in range(
            max(0, i - radius), min(max_index, i + radius + 1)
        )
    ]
    return set(new_list)



import re
from io import StringIO
from Bio import Restriction
from Bio.Seq import Seq
from Bio import SeqIO
from .SequenceString import SequenceString
import numpy as np

def get_sequence_topology(sequence, default="linear"):
    if isinstance(sequence, SequenceString):
        return sequence.metadata.get("topology", default)
    else:
        return default


def gc_content(sequence, window_size=None):
    """Compute global or local GC content.

    Parameters
    ----------

    sequence
      An ATGC DNA sequence (upper case!)

    window_size
      If provided, the local GC content for the different sliding windows of
      this size is returned, else the global GC content is returned.

    Returns
    --------

      A number between 0 and 1 indication the proportion
      of GC content. If window_size is provided, returns
      a list of len(sequence)-window_size values indicating
      the local GC contents (sliding-window method). The i-th value
      indicates the GC content in the window [i, i+window_size]
    """
    # The code is a little cryptic but speed gain is 300x
    # compared with pure-python string operations

    arr = np.fromstring(sequence + "", dtype="uint8")
    arr_GCs = (arr == 71) | (arr == 67)  # 67=C, 71=G

    if window_size is None:
        return 1.0 * arr_GCs.sum() / len(sequence)
    else:
        cs = np.cumsum(arr_GCs)
        a = cs[window_size - 1 :]
        b = np.hstack([[0], cs[:-window_size]])
        return 1.0 * (a - b) / window_size


def find_enzyme_sites(
    sequence, enzyme_name, padding=0, padding_nuc="A"
):
    padding = padding * padding_nuc
    if hasattr(sequence, 'seq'):
        sequence = str(sequence.seq)
    sequence = str(sequence)
    sequence = Seq(padding + sequence + padding)
    topology = get_sequence_topology(sequence, default="linear")
    linear = (topology == "linear")
    return Restriction.__dict__[enzyme_name].search(sequence, linear=linear)

def gc_content_to_tm(seq_length, gc_content, gc_tm=4, nongc_tm=2):
    return seq_length * (gc_content * gc_tm + (1 - gc_content) * nongc_tm)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Path


def plot_tree_graph(
    levels,
    edges,
    draw_node,
    elements_positions=None,
    ax=None,
    width_factor=2.5,
    height_factor=2,
    scale=1.0,
    edge_left_space=0.015,
    edge_right_space=0.015,
    interlevel_shift=0,
    margin=None,
    **txt_kw
):
    """General function for plotting tree graphs.

    Parameters
    ----------

    levels
      A list of lists of nodes grouped by "level", i.e distance to the in the
      graph to the level 0. levels will be displayed on a same column.

    edges
      List of nodes pairs (source node, target node).

    draw_node
      A function f(x , y , node, ax, **kw) which draws something related to the
      node at the position x,y on the given Matplotlib ax.

    ax
      The matplotlib ax to use. If none is provided, a new ax is generated.

    Returns
    -------
    elements_positions, ax
      Dictionary of elements positions, matplotlib ax.

    Examples:
    ---------

    >>> def draw_node(x,y, node, ax):
        ax.text(x,y, node)
    >>> plot_tree_graph(levels=[["A","B","C"], ["D,E"], ["F"]],
                        edges=[("A","D"),("B","D"),("C","E")
                               ("D","F"),("E","F")],
                        draw_node = draw_node,)



    """
    levels_dict = {
        element: level
        for level, elements in enumerate(levels)
        for element in elements
    }
    if elements_positions is None:
        elements_positions = {}
        for lvl, elements in enumerate(levels):
            yy = np.linspace(0, 1, len(elements) + 2)[1:-1]
            yy += interlevel_shift * (1 - 2 * (lvl % 2))
            x = 1.0 * (1 + lvl) / (len(levels) + 1)
            for y, element in zip(yy, elements):
                elements_positions[element] = (x, y)

    if ax is None:
        width = width_factor * len(levels) * scale
        height = height_factor * max([len(lvl) for lvl in levels]) * scale
        fig, ax = plt.subplots(1, figsize=(width, height))

    for element, (x, y) in elements_positions.items():
        draw_node(x, y, element, ax, **txt_kw)

    y_spans = [
        elements_positions[elements[1]][1] - elements_positions[elements[0]][1]
        for elements in levels
        if len(elements) > 1
    ]

    delta_y = 0.5 * min(y_spans) if y_spans != [] else 0

    for el1, el2 in edges:
        x1, y1 = elements_positions[el1]
        x2, y2 = elements_positions[el2]
        x1 += edge_left_space * np.sqrt(scale)
        x2 += -edge_right_space * np.sqrt(scale)
        if ((levels_dict[el2] - levels_dict[el1]) > 1) and (y1 == y2):
            patch = mpatches.PathPatch(
                Path(
                    [
                        (x1, y1),
                        (0.5 * x2 + 0.5 * x1, y1 - delta_y),
                        (0.5 * x2 + 0.5 * x1, y2 - delta_y),
                        (x2, y2),
                    ],
                    [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4],
                ),
                facecolor="none",
                lw=1 * scale,
                zorder=-1000,
            )

        else:
            patch = mpatches.PathPatch(
                Path(
                    [
                        (x1, y1),
                        (0.9 * x2 + 0.1 * x1, y1),
                        (0.1 * x2 + 0.9 * x1, y2),
                        (x2, y2),
                    ],
                    [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4],
                ),
                facecolor="none",
                lw=1 * scale,
                zorder=-1000,
            )

        ax.add_patch(patch)

    ax.axis("off")
    if margin is not None:
        xx, yy = [np.array(e) for e in zip(*elements_positions.values())]
        xmin, xmax = xx.min(), xx.max()
        dx = margin * (xmax - xmin)
        ymin, ymax = yy.min(), yy.max()
        dy = margin * (ymax - ymin)
        d = max(dx, dy)
        ax.set_xlim(xmin - d, xmax + d)
        ax.set_ylim(ymin - d, ymax + d)

    return elements_positions, ax



import os
import datetime
import weasyprint
import pandas
from jinja2 import Environment, FileSystemLoader
from ..config import SETTINGS
from .PlotsMixin.give_quotes_html_locations import give_quotes_html_locations


class PdfReportMixin:
    def make_html_report(self):
        """Return a HTML version of the assembly report (later converted to PDF)"""

        def _anchor_span(op):
            return "<span id='anchor_%s'>%s</span>" % (op.id, op.id)

        operations = self.to_steps_list()

        def sorting_key(op):
            location = op.get("final_location", None)
            if location is None:
                location = (-1, 1)
            return (location[0], -location[1])

        operations = sorted(operations, key=sorting_key)

        give_quotes_html_locations(
            operations, len_sequence=len(self.plan.sequence)
        )
        env = Environment(loader=FileSystemLoader(SETTINGS["template_path"]))
        template = env.get_template("report_template.html")

        orders_dataframe = pandas.DataFrame.from_records(
            [
                {
                    "ID": _anchor_span(op),
                    "Company": op.source,
                    "Sequence": op.sequence,
                    "Length": len(op.sequence),
                    "Price": op.price,
                    "Lead time": op.lead_time,
                    "Deadline": op.get("deadline", None),
                    "Location": op.html_location,
                }
                for op in operations
                if self.sources[op.source].operation_type == "order"
            ]
        )

        pcrs_dataframe = pandas.DataFrame.from_records(
            [
                {
                    "ID": _anchor_span(op),
                    "Primers": ", ".join(
                        [
                            _anchor_span(child)
                            for child in op.get("assembly_plan", [])
                        ]
                    ),
                    "Source": op.source,
                    "Sequence": op.sequence,
                    "Length": len(op.sequence),
                    "Price": op.price,
                    "Lead time": op.lead_time,
                    "Deadline": op.get("deadline", None),
                    "Location": op.html_location,
                }
                for op in operations
                if self.sources[op.source].operation_type == "PCR"
            ],
            columns=[
                "ID",
                "Source",
                "infos",
                "Primers",
                "Sequence",
                "Length",
                "Price",
                "Lead time",
                "Location",
            ],
        )

        parts_dataframe = pandas.DataFrame.from_records(
            [
                {
                    "ID": _anchor_span(op),
                    "Source": op.source,
                    "Sequence": op.sequence,
                    "Length": len(op.sequence),
                    "Price": op.price,
                    "Lead time": op.lead_time,
                    "Deadline": op.get("deadline", None),
                    "Location": op.html_location,
                }
                for op in operations
                if self.sources[op.source].operation_type == "library"
            ],
            columns=[
                "ID",
                "Source",
                "Sequence",
                "Length",
                "Price",
                "Lead time",
                "Location",
            ],
        )

        assembly_operations = [
            op
            for op in operations
            if self.sources[op.source].operation_type == "assembly"
        ]

        asm_dataframe = pandas.DataFrame.from_records(
            [
                {
                    "ID": _anchor_span(op),
                    "Station": op.source,
                    "Sequence": op.sequence,
                    "Length": len(op.sequence),
                    "Price": op.price,
                    "Lead time": op.lead_time,
                    "Deadline": op.get("deadline", None),
                    "Location": op.html_location,
                    "Primers": None,
                }
                for op in assembly_operations
            ],
            columns=[
                "ID",
                "Station",
                "infos",
                "Sequence",
                "Length",
                "Price",
                "Lead time",
                "Location",
            ],
        )

        cost_orders = orders_dataframe["Price"].sum()

        render_parameters = {
            "sequence": self.plan.sequence,
            "sequence_name": self.plan.get("sequence_name", "unknown"),
            "sequence_hash": "%02x" % hash(self.plan.sequence),
            "quote": self.plan,
            "date_submitted": datetime.datetime.now(),
            "n_orders": len(orders_dataframe),
            "cost_orders": cost_orders,
            "total_cost": self.plan.price,
            "orders": orders_dataframe,
            "asm_dataframe": asm_dataframe,
            "assembly_operations": assembly_operations,
            "user_name": self.plan.get("user_name", "unknown"),
        }
        if len(pcrs_dataframe) > 0:
            render_parameters["pcr_reuses"] = pcrs_dataframe

        if len(parts_dataframe) > 0:
            render_parameters["parts_reuses"] = parts_dataframe

        html_out = template.render(render_parameters)

        return html_out

    def write_pdf_report(self, target):
        """Return a PDF version of the report with general infos and details of
        each intermediary constructs."""
        base_path = SETTINGS["template_path"]
        css_path = os.path.join(base_path, "static", "css")
        stylesheets = [
            os.path.join(css_path, "dnaweaver_report.css"),
            # os.path.join(css_path, "font-awesome.min.css")
        ]
        html_out = self.make_html_report()
        weasy_html = weasyprint.HTML(string=html_out, base_url=base_path)
        weasy_html.write_pdf(target, stylesheets=stylesheets)



import colorsys
import itertools
import matplotlib.colors as cl
import matplotlib.cm as cm

def hls_to_hex(hue, luminance, saturation):
    """Return (R,G,B) equivalent of a hue/staturation/value color."""
    return cl.rgb2hex(colorsys.hls_to_rgb(hue, luminance, saturation))


def rgb_to_hex(red, green, blue):
    """Return color as #rrggbb for the given color values."""
    return "#%02x%02x%02x" % (int(red), int(green), int(blue))

class ColorsMixin:


    def autocolor_quote_sources(
        self,
        hues=(0.635, 0.047, 0.117),
        saturations=(0.9, 0.7, 0.5, 0.3),
        min_lum=0.2,
        max_lum=0.8,
    ):
        """Auto-add a `_report_color` field to the sources in in quote.sources.

        Sources at the same depth share the same luminance
        """

        colors = itertools.cycle(
            [
                rgb_to_hex(
                    *[255 * e ** 0.4 for e in cm.Paired(0.13 * i % 1.0)][:3]
                )
                for i in range(30)
            ]
        )
        for _name, source in sorted(self.sources.items()):
            color = next(colors)
            source._report_color = color



""" This module contains overhangs selectors for methods such as
Gibson Assembly, Golden Gate assembly, recombination in yeast."""

from ..biotools import get_sequence_topology

class SegmentSelector:
    """Base class for segment selectors such as TmSegmentSelector.

    These selectors return an ideal segment subsegment around a given sequence
    location. They can also be used to filter out locations in the sequence as
    non-potential cutting sites for sequence decomposition.
    """

    def location_filter_method(self, sequence):
        """Return a filter function f(location) => True/False.

        The result is True iff the location is a valid cutting site"""
        if self.has_location_filter:

            def f(index):
                return self.filter_location(sequence, index)

            return f
        else:
            return None

    @property
    def has_location_filter(self):
        return hasattr(self, "filter_location")

    def compute_fragment_for_sequence_segment(self, sequence, segment):
        """Compute the fragment as (sequence region + flank segments)."""
        start, end = segment
        topology = get_sequence_topology(sequence)
        if topology == "circular":
            half_homology = self.max_homology_size // 2
            if start <= half_homology:
                return self.compute_fragment_for_sequence_segment(
                    sequence=sequence[-half_homology:] + sequence,
                    segment=(start + half_homology, end + half_homology),
                )
            if end >= len(sequence) - half_homology:
                return self.compute_fragment_for_sequence_segment(
                    sequence=sequence + sequence[:half_homology],
                    segment=segment,
                )
        if start == 0:
            fragment_start = 0
        else:
            loc = self.compute_segment_location(sequence, start)
            if loc is None:
                subseq = sequence[:20] + "..."
                raise ValueError(
                    "loc was None around start %s for sequence %s (%dbp)"
                    % (end, subseq, len(sequence))
                )
            fragment_start = loc[0]
        if end == len(sequence):
            fragment_end = len(sequence)
        else:
            loc = self.compute_segment_location(sequence, end)
            if loc is None:
                subseq = sequence[:20] + "..."
                raise ValueError(
                    "loc was None around end %s for sequence %s (%dbp)"
                    % (end, subseq, len(sequence))
                )
            fragment_end = loc[1]
        fragment = sequence[fragment_start:fragment_end]
        return self.left_addition + fragment + self.right_addition

    def compute_segment_around_index(self, sequence, index, topology="linear"):
        """Return the sequence of the selected segment at the given index."""

        if get_sequence_topology(sequence) == "circular":
            half_homology = self.max_homology_size // 2
            if index <= half_homology:
                return self.compute_segment_around_index(
                    sequence=sequence[-half_homology:] + sequence,
                    index=index + half_homology,
                )
            if index >= len(sequence) - half_homology:
                return self.compute_segment_around_index(
                    sequence=sequence + sequence[:half_homology], index=index,
                )
        start, end = self.compute_segment_location(sequence, index)
        return sequence[start:end]

    @staticmethod
    def get_segment_coordinates(center, segment_length, sequence_length):
        """Return max(0, c - s/2) - min(L, c + L/2).

        Where c=center, s=segment_length, L=sequence_length.
        """
        half = int(segment_length / 2)
        start = max(0, min(center - half, sequence_length - segment_length))
        end = start + segment_length
        return start, end

    def __repr__(self):
        return str(self)



import itertools
from ..biotools import (
    reverse_complement,
    gc_content_to_tm,
    find_enzyme_sites,
    get_sequence_topology,
)

from ..tools import memoize
from ..SegmentSelector import TmSegmentSelector
from .OverlapingAssemblyMethod import OverlapingAssemblyMethod


class GoldenGateAssemblyMethod(OverlapingAssemblyMethod):
    """The Golden Gate Assembly Method.

    This method adds overhangs with a Type IIS REase site to the segments.

    Parameters
    ----------

    left_overhang
      An ATGC DNA sequence representing the left overhang, in 5' -3'.
      For practicality, this can contain "[BsaI]", "[BsmBI]", "[BbsI]", which
      will be replaced by their restriction sites


    right_overhang
      An ATGC DNA sequence representing the right overhang, in 5'-3'.
      Be careful, I said 5'-3' !!!
      If left to None, will be equal to left_overhang.

    """

    name = "Golden Gate Assembly"

    enzymes_dict = {"BsaI": "GGTCTC", "BsmBI": "CGTCTC", "BbsI": "GAAGAC"}

    def __init__(
        self,
        enzyme="BsaI",
        wildcard_basepair="A",
        left_addition="",
        right_addition="",
        refuse_sequences_with_enzyme_site=True,
        min_overhangs_gc=0,
        max_overhangs_gc=1,
        min_overhangs_differences=1,
        **props
    ):
        if enzyme not in self.enzymes_dict:
            raise ValueError(
                "Enzyme should be one of %s" % self.enzymes_dict.keys()
            )

        self.min_overhangs_gc = min_gc = min_overhangs_gc
        self.max_overhangs_gc = max_gc = max_overhangs_gc
        self.min_overhangs_differences = min_overhangs_differences

        self.enzyme = enzyme
        self.enzyme_site = self.enzymes_dict[enzyme]
        enzyme_site_plus_basepair = self.enzyme_site + wildcard_basepair
        self.left_addition = left_addition + enzyme_site_plus_basepair
        self.right_addition = (
            reverse_complement(enzyme_site_plus_basepair) + right_addition
        )
        self.refuse_sequences_with_enzyme_site = (
            refuse_sequences_with_enzyme_site
        )

        overhang_selector = TmSegmentSelector(
            min_size=4,
            max_size=4,
            min_tm=gc_content_to_tm(4, min_gc),
            max_tm=gc_content_to_tm(4, max_gc),
            left_addition=self.left_addition,
            right_addition=self.right_addition,
        )
        OverlapingAssemblyMethod.__init__(self, overhang_selector, **props)

        # CUTS LOCATION CONSTRAINT BASED ON GC CONTENT

        if refuse_sequences_with_enzyme_site:

            def no_site_in_sequence(sequence):
                sites = find_enzyme_sites(sequence, enzyme_name=self.enzyme)
                return sites == []

            self.sequence_constraints.append(no_site_in_sequence)

        # DO NOT CUT AT PALYINDROMIC REGIONS

        def no_cut_at_palyndromic_locations(sequence):
            L = len(sequence)

            def no_palyndrom_filter(i):
                s = overhang_selector.compute_segment_around_index(sequence, i)
                rev_s = reverse_complement(s)
                rev_diffs = len([a for a, b in zip(s, rev_s) if a != b])
                assert len(s) == 4
                return rev_diffs >= self.min_overhangs_differences

            return no_palyndrom_filter

        self.cut_location_constraints.append(no_cut_at_palyndromic_locations)

        # CUTS SET CONSTRAINT: ALL OVERHANGS MUST BE COMPATIBLE

        def overhangs_are_compatible(o1, o2):
            diffs = len([a for a, b in zip(o1, o2) if a != b])
            if diffs >= self.min_overhangs_differences:
                rev_o2 = reverse_complement(o2)
                rev_diffs = len([a for a, b in zip(o1, rev_o2) if a != b])
                return rev_diffs >= self.min_overhangs_differences
            return False

        overhangs_are_compatible = memoize(overhangs_are_compatible)

        def all_overhangs_are_compatible(sequence):
            topology = get_sequence_topology(sequence, "linear")

            def constraint(cut_locations):
                cut_overhangs = {
                    cut_location: overhang_selector.compute_segment_around_index(
                        sequence, cut_location
                    )
                    for cut_location in cut_locations
                }
                cut_pairs = list(itertools.combinations(cut_locations, 2))
                if topology == "circular":
                    cut_pairs.remove((0, len(sequence)))

                return all(
                    [
                        overhangs_are_compatible(
                            cut_overhangs[c1], cut_overhangs[c2]
                        )
                        for c1, c2 in cut_pairs
                    ]
                )

                # overhangs = sorted(
                #     [
                #         overhang_selector.compute_segment_around_index(
                #             sequence, cut_location
                #         )
                #         for cut_location in cut_locations
                #     ]
                # )
                # return all(
                #     [
                #         overhangs_are_compatible(o1, o2)
                #         for o1, o2 in itertools.combinations(overhangs, 2)
                #     ]
                # )

            return constraint

        self.cuts_set_constraints.append(all_overhangs_are_compatible)

    def additional_dict_description(self):
        return {
            "enzyme": self.enzyme,
            "left addition": self.left_addition,
            "right addition": self.right_addition,
            "refuse sequences with enzyme site": str(
                self.refuse_sequences_with_enzyme_site
            ),
            "overhangs gc content": "%d-%d%%"
            % (100 * self.min_overhangs_gc, 100 * self.max_overhangs_gc),
            "overhangs differences": self.min_overhangs_differences,
        }



from .DnaAssemblyMethod import DnaAssemblyMethod
from .GoldenGateAssemblyMethod import GoldenGateAssemblyMethod
from .OverlapingAssemblyMethod import (
    GibsonAssemblyMethod,
    BuildAGenomeAssemblyMethod,
)
from .BluntEndAssemblyMethod import BluntEndAssemblyMethod

__all__ = [
    "DnaAssemblyMethod",
    "GoldenGateAssemblyMethod",
    "GibsonAssemblyMethod",
    "BuildAGenomeAssemblyMethod",
    "BluntEndAssemblyMethod",
]




import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import numpy as np

from ...config import SETTINGS

class TimelineMixin:


    def plot_assembly_timeline(
        self, deadline=None, ax=None, rectangle_color="#bbbbff", scale=1.0
    ):
        """Plot the assembly timeline with Matplotlib."""
        if deadline is None:
            deadline = self.plan.lead_time
        self.propagate_deadline(deadline)
        assemblies_list = self.tree_as_list()
        assemblies_list = sorted(assemblies_list, key=lambda a: a.id)

        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12 * scale,
            family="sans-serif",
        )
        fontawesome = fm.FontProperties(
            fname=SETTINGS["fontawesome-ttf-path"],
            size=12 * scale,
            family="sans-serif",
        )
        if ax is None:
            _fig, ax = plt.subplots(
                1, figsize=(16 * scale, len(assemblies_list) * 0.3 * scale)
            )
        positions = {}

        if textprops is None:
            textprops = fm.FontProperties(
                fname=SETTINGS["OpenSans-ttf-path"], size=12, family="sans-serif"
            )

        for i, assembly in enumerate(assemblies_list):
            source = assembly.source
            color = "k"
            if source.__class__.__name__ == "CommercialDnaOffer":
                color = "#888888"
            x2 = assembly.deadline
            x1 = assembly.deadline - assembly.step_duration
            positions[assembly.id] = ([x1, i], [x2, i])
            patch = mpatches.Rectangle(
                (assembly.deadline - assembly.step_duration, i - 0.2),
                assembly.step_duration,
                0.4,
                color=rectangle_color,
                ec="none",
            )

            ax.plot(
                [0, assembly.deadline],
                [i, i],
                c="k",
                alpha=0.3,
                lw=0.5 * scale,
                zorder=-1000,
                ls="-",
            )
            ax.add_patch(patch)

            ax.text(
                -1.5,
                i,
                assembly.id,
                horizontalalignment="right",
                verticalalignment="center",
                fontproperties=textprops,
                color=color,
            )
            ax.text(
                -0.75,
                i,
                source.report_fa_symbol,
                horizontalalignment="center",
                verticalalignment="center",
                fontproperties=fontawesome,
                color=color,
            )

        for assembly in assemblies_list:
            xa, ya = positions[assembly.id][0]
            if assembly.assembly_plan is not None:
                for _segment, child in assembly.assembly_plan.items():
                    xc, yc = positions[child.id][1]
                    ax.plot([xc, xa], [yc, ya], lw=1 * scale, color=color)

        ax.set_ylim((-1, len(assemblies_list) + 1))
        ax.set_xlim(-2, deadline)
        ax.set_frame_on(False)
        ax.yaxis.set_visible(False)
        ticks = list(np.arange(0, deadline, deadline / 6).astype(int))

        ax.xaxis.set_ticks(ticks)
        ax.xaxis.set_ticks_position("bottom")
        ax.xaxis.set_tick_params(which="major", length=5 * scale)
        ax.xaxis.set_ticklabels(ticks, fontproperties=textprops)
        ax.set_xlabel("TIME", fontproperties=textprops)
        ax.set_frame_on(False)
        return ax



"""Nice function to make figures for slides (not used in the framework)"""

import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

def plot_decomposition_graph(
    graph,
    nodes_color="#6886b7",
    weight="weight",
    colormap="jet",
    ax=None,
    edge_width=1,
    edges_step=1,
    edge_alpha=0.2,
    figsize=(8, 8),
):
    xx = np.array(sorted(graph.nodes))
    L = xx.max()
    edges = [
        (start, end, data[weight] / (end - start))
        for start, end, data in graph.edges(data=True)
    ]
    edges = edges[::edges_step]
    edges = sorted(edges, key=lambda e: e[2])
    max_segment_length = max([end - start for (start, end, _) in edges])
    weights = np.array([w for (_, _, w) in edges])
    normalized_weights = (255 * weights / weights.max()).astype("uint8")
    colormap = cm.__dict__[colormap]
    colors = colormap(normalized_weights, alpha=edge_alpha)
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    ax.axis("off")
    #     print (list(colors))
    for (start, end, w), color in zip(edges, colors):
        xc = 0.5 * (start + end)
        half = 0.5 * abs(end - start)
        ax.add_patch(
            mpatches.Arc(
                (xc, 0),
                2 * half,
                2 * half,
                theta1=0,
                theta2=180,
                facecolor="none",
                ls="-",
                edgecolor=color,
                linewidth=1,
            )
        )
    ax.plot(xx, [0 for x in xx], marker="o", c=nodes_color)
    ax.set_aspect("equal")
    ax.set_xlim(-1, L + 1)
    ax.set_ylim(-1, max_segment_length / 2 + 1)





from .sequence_homologies import (
    blast_sequence,
    perfect_match_locations_in_hsp,
    largest_common_substring,
)

from .sequence_operations import (
    random_dna_sequence,
    reverse_complement,
    string_to_sequence,
    load_record,
)
from .sequence_analysis import (
    gc_content,
    gc_content_to_tm,
    find_enzyme_sites,
    get_sequence_topology
)
from .SequenceString import SequenceString


import numpy as np
from ....DnaQuote import DnaQuote
from ....SegmentSelector import TmSegmentSelector, FixedSizeSegmentSelector
from ....DnaAssemblyMethod import (
    BuildAGenomeAssemblyMethod,
    GibsonAssemblyMethod,
    GoldenGateAssemblyMethod,
)
from ...builtin_constraints import SequenceLengthConstraint
from ...DnaSupplier import DnaSupplier
from ..DnaSuppliersComparator import DnaSuppliersComparator
from .SequenceDecomposer import SequenceDecomposer, NoSolutionFoundError



class DnaAssemblyStation(DnaSupplier):
    """DNA Assembly stations assemble together DNA fragments using a specific
    assembly method.

    Parameters
    ----------

    name
      Name of the station (appears on reports)

    assembly_method
      AnDnaAssemblyMethod object specifying how the fragments are assembled,
      what sequences can be assembled, what fragments can be used, etc.

    supplier

    """

    class_description = "DNA assembly station"
    operation_type = "assembly"
    report_fa_symbol = u""
    report_fa_symbol_plain = "flask"
    report_color = "#eeeeff"

    def __init__(
        self,
        name,
        assembly_method,
        supplier,
        memoize=False,
        decomposer_class=None,
        a_star_auto_multiplier=2,
        **solver_kwargs
    ):
        self.name = name
        self.assembly_method = assembly_method
        self.set_suppliers(supplier)
        self.extra_time = assembly_method.duration
        self.extra_cost = assembly_method.cost
        self.sequence_constraints = assembly_method.sequence_constraints
        self.cuts_set_constraints = assembly_method.cuts_set_constraints
        self.memoize = memoize
        if decomposer_class is None:
            self.decomposer_class = SequenceDecomposer
        else:
            self.decomposer_class = decomposer_class
        self.memoize_dict = {}
        self.min_basepair_price = self.supplier.min_basepair_price
        if solver_kwargs.get("a_star_factor", None) == "auto":
            solver_kwargs["a_star_factor"] = (
                a_star_auto_multiplier * self.min_basepair_price
            )
        self.solver_kwargs = solver_kwargs

    def get_quote_for_sequence_segment(
        self, sequence, segment, max_lead_time=None, **kwargs
    ):
        """Return the cost of the segment

        Is used as the "cost" function for a segment during decomposition
        optimization

        """
        fragment_to_order = self.assembly_method.compute_fragment_for_sequence_segment(
            sequence=sequence, segment=segment, **kwargs
        )
        return self.supplier.get_quote(
            fragment_to_order, max_lead_time=max_lead_time
        )

    def get_assembly_plan_from_cuts(self, sequence, cuts, max_lead_time=None):
        """Return a plan {(seg,ment): quote, ...} based on the cut positions.
        """
        cuts = sorted(cuts)
        return {
            segment: self.get_quote_for_sequence_segment(
                sequence, segment, max_lead_time=max_lead_time
            )
            for segment in zip(cuts, cuts[1:])
        }

    def new_sequence_decomposer(
        self,
        sequence,
        max_lead_time=None,
        coarse_grain=1,
        fine_grain=1,
        cut_spread_radius=0,
        a_star_factor=0,
        logger=None,
    ):
        def segment_score(segment):
            quote = self.get_quote_for_sequence_segment(
                sequence, segment, max_lead_time=max_lead_time,
            )
            if not quote.accepted:
                return -1
            else:
                return quote.price

        assembly = self.assembly_method

        return self.decomposer_class(
            sequence_length=len(sequence),
            segment_score_function=segment_score,
            cut_location_constraints=[
                cs(sequence) for cs in assembly.cut_location_constraints
            ],
            segment_constraints=[
                cs(sequence) for cs in assembly.segment_constraints
            ],
            min_segment_length=assembly.min_segment_length,
            max_segment_length=assembly.max_segment_length,
            forced_cuts=assembly.force_cuts(sequence),
            suggested_cuts=self.compute_suggested_cuts(sequence),
            coarse_grain=coarse_grain,
            fine_grain=fine_grain,
            logger=logger,
            bar_prefix="[%s] " % self.name,
            a_star_factor=a_star_factor,
            path_size_limit=assembly.max_fragments,
            cuts_set_constraints=[
                cs(sequence) for cs in assembly.cuts_set_constraints
            ],
            cut_spread_radius=cut_spread_radius,
        )

    def compute_suggested_cuts(self, sequence):
        all_supplier_cuts = []
        for supplier in self.suppliers:
            if hasattr(supplier, "suggest_cuts"):
                supplier_cuts = supplier.suggest_cuts(sequence)
                all_supplier_cuts.extend(list(supplier_cuts))
        assembly_method_cuts = self.assembly_method.suggest_cuts(sequence)
        suggested = list(assembly_method_cuts) + list(all_supplier_cuts)
        return sorted(set(suggested))

    def suggest_cuts(self, sequence):
        return self.compute_suggested_cuts(sequence=sequence)

    def get_assembly_plan_for_sequence(
        self,
        sequence,
        max_lead_time=None,
        coarse_grain=1,
        fine_grain=1,
        cut_spread_radius=0,
        a_star_factor=0,
        logger=None,
    ):
        """Return the plan {(seg, ment): quote, ...} of the optimal strategy
        for the sequence's decomposition."""
        # extended_sequence = self.assembly_method.extend_sequence(sequence)
        decomposer = self.new_sequence_decomposer(
            sequence=sequence,
            max_lead_time=max_lead_time,
            coarse_grain=coarse_grain,
            fine_grain=fine_grain,
            a_star_factor=a_star_factor,
            cut_spread_radius=cut_spread_radius,
            logger=logger,
        )
        self._lattest_decomposer = decomposer  # for debugging
        best_cuts = decomposer.compute_optimal_cuts()
        return self.get_assembly_plan_from_cuts(
            sequence, best_cuts, max_lead_time=max_lead_time,
        )

    def get_best_price(
        self, sequence, max_lead_time=None, with_assembly_plan=False,
    ):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        if max_lead_time is not None:
            max_lead_time = max_lead_time - self.extra_time
            if max_lead_time < 0:
                return DnaQuote(
                    self,
                    sequence,
                    accepted=False,
                    message="Lead time limit too short",
                )
        try:
            assembly_plan = self.get_assembly_plan_for_sequence(
                sequence, max_lead_time=max_lead_time, **self.solver_kwargs
            )
        except NoSolutionFoundError as err:
            message = "No solution found! %s" % err
            print (message)
            return DnaQuote(self, sequence, accepted=False, message=message)

        # A solution has been found ! Now compute overall time and lead time.

        if any([q.price is None for q in assembly_plan.values()]):
            return DnaQuote(
                self, sequence, accepted=False, message="No solution found !"
            )
        quote = DnaQuote(self, sequence, assembly_plan=assembly_plan)
        quote.price = quote.children_total_price() + self.extra_cost
        children_lead_time = quote.children_overall_lead_time()
        if (children_lead_time is None) or (self.extra_time is None):
            quote.lead_time = None
        else:
            quote.lead_time = children_lead_time + self.extra_time
        if not with_assembly_plan:
            self.assembly_plan = None
        return quote

    def additional_dict_description(self):
        result = {
            "dna source": self.supplier.name,
            "solver parameters": self.solver_kwargs,
        }
        result.update(
            {
                ("assembly method %s" % k): v
                for k, v in self.assembly_method.dict_description().items()
            }
        )
        return result

    @staticmethod
    def from_dict(data):
        sequence_constraints = []
        if data["use_size_range"]:
            mini, maxi = data["size_range"]
            sequence_constraints.append(SequenceLengthConstraint(mini, maxi))
        min_length, max_length = data["fragments_size_range"]
        if data["method"] == "type_iis":
            gc_range = data.get("overhang_gc_range", [0, 1])
            method = GoldenGateAssemblyMethod(
                duration=data["duration"],
                cost=data["cost"],
                min_overhangs_gc=gc_range[0],
                max_overhangs_gc=gc_range[1],
                enzyme=data["enzyme"],
                max_segment_length=max_length,
                min_segment_length=min_length,
                max_fragments=data["max_fragments"],
                sequence_constraints=sequence_constraints,
            )
        elif data["method"] in [
            "gibson_assembly",
            "yeast_recombination",
            "oligo_assembly",
        ]:
            if data["overhang_type"] == "tm":
                min_oh_size, max_oh_size = data["overhang_size_range"]
                min_tm, max_tm = data["tm_range"]
                overhang_selector = TmSegmentSelector(
                    min_size=min_oh_size,
                    max_size=max_oh_size,
                    min_tm=min_tm,
                    max_tm=max_tm,
                )
            else:
                overhang_selector = FixedSizeSegmentSelector(
                    segment_size=data["overlap"]
                )
            if data["method"] == "oligo_assembly":
                method_class = BuildAGenomeAssemblyMethod
            else:
                method_class = GibsonAssemblyMethod
            method = method_class(
                duration=data["duration"],
                cost=data["cost"],
                overhang_selector=overhang_selector,
                max_segment_length=max_length,
                min_segment_length=min_length,
                max_fragments=data["max_fragments"],
                sequence_constraints=sequence_constraints,
            )
        if data["use_astar"]:
            a_star_factor = data.get("astar_factor", "auto")
        else:
            a_star_factor = 0
        max_construct_length = max_length * data["max_fragments"]
        if data["grain_type"] == "auto":
            data["coarse_grain"] = int(max_construct_length / 15)
            data["fine_grain"] = int(np.sqrt(data["coarse_grain"]))

        return DnaAssemblyStation(
            name=data["name"],
            supplier=data["suppliers"],
            assembly_method=method,
            coarse_grain=data["coarse_grain"],
            fine_grain=data["fine_grain"],
            a_star_factor=a_star_factor,
            memoize=data.get("memoize", False),
            logger=data.get("logger", None),
        )

    def set_suppliers(self, suppliers):
        if hasattr(suppliers, "__iter__"):
            if len(suppliers) > 1:
                self.supplier = DnaSuppliersComparator(
                    name=self.name + " comparator", suppliers=suppliers
                )
                self.supplier.is_ghost_source = True
            else:
                self.supplier = suppliers[0]
        else:
            self.supplier = suppliers
            suppliers = [suppliers]
        self.suppliers = suppliers



from dna_features_viewer import GraphicRecord
import matplotlib.pyplot as plt
from copy import deepcopy
import matplotlib.patches as mpatches
import matplotlib.font_manager as fm
from Bio import SeqIO
from ...config import SETTINGS

class AssemblyBlocksMixin:
    
    def plot_assembly_blocks(
        self,
        parts_offset=0,
        backend="matplotlib",
        plot_top_assembly=True,
        ax=None,
        edge_widths=None,
        legend=False,
        legend_offset=-0.05
    ):
        """Return a Matplotlib or Bokeh plot of the assembly tree of blocks.

        Parameters
        ----------



        parts_offset
        Offset applied so that consecutive blocks are not exactly on the same
        level. Can go from 0 (flat line of blocks) to e.g. 1

        backend
        One of "matplotlib" or "bokeh". Determines the library used to produce
        the graph. matplotlib can output PNG/SVG/etc. while Bokeh outputs an
        interactive HTML/JS graph

        plot_top_assembly
        Whether the top assembly (which is just one big block) should be plotted
        or not

        ax
        A Matplotlib Axes object. If no ax is provided, a new figure and ax are
        generated.

        edge_widths
        A dict {sourcename : width} indicating the widths of the rectangles based
        on the source name. Rectangles with a very small width will be edgeless.

        legend
        Whether the legend is included in the ax.

        """
        rectangles = []
        if edge_widths is None:
            edge_widths = {}
        if not hasattr(self, "plan"):
            self.compute_full_assembly_plan()
        tree = deepcopy(self.plan)

        def rec(_quote, depth=0, xstart=0, xend=None):
            children = _quote.pop("assembly_plan")
            # print _quote.final_location
            if _quote.final_location is not None:
                left, right = _quote.final_location
            else:
                left, right = xstart, xend

            if children is None:
                children = []
            if plot_top_assembly or (depth > 0):
                source = self.sources[_quote.source]
                rectangles.append(
                    {
                        "top": -depth + 0.5 + (not plot_top_assembly),
                        "bottom": -depth + (not plot_top_assembly),
                        "left": left,
                        "right": right,
                        "source": str(_quote.source),
                        "lead_time": str(_quote.lead_time),
                        "price": str(_quote.price),
                        "length": len(_quote.sequence),
                        "has_children": children != [],
                        "line_width": 1,
                        "operation_type": source.operation_type,
                        "fa_symbol": source._report_fa_symbol_plain,
                        "color": source._report_color,
                    }
                )
            for child in children:
                rec(
                    child,
                    depth=depth + 1,
                    xstart=xstart + child.segment_start,
                    xend=xstart + child.segment_end,
                )

        rec(tree, xend=len(tree.sequence))
        tops = list(set([r["top"] for r in rectangles]))
        offsets = {top: 0 for top in tops}

        guides = []
        for top in sorted(tops):
            rects = sorted(
                [r for r in rectangles if r["top"] == top], key=lambda r: r["left"]
            )
            for i, rect in enumerate(rects):
                color = ["#e5ecff", "#ffffe5"][i % 2]
                if rect["has_children"]:
                    guides.append(
                        {
                            "top": rect["top"],
                            "bottom": rect["top"] - 1,
                            "left": rect["left"],
                            "right": rect["right"],
                            "color": color,
                        }
                    )
        for rectangle in sorted(rectangles, key=lambda r: r["left"]):
            if parts_offset:
                offset = offsets[rectangle["top"]]
                offsets[rectangle["top"]] = (offsets[rectangle["top"]] + 1) % 2
                rectangle["top"] += parts_offset * offset
                rectangle["bottom"] += parts_offset * offset

        if hasattr(self, "genbank") and (self.genbank is not None):
            record = SeqIO.read(self.genbank, "genbank")
        else:
            record = None
        return _matplotlib_plot_assembly_blocks(
            rectangles, guides, tops, ax, record=record, legend=legend,
            legend_offset=legend_offset
        )


def _matplotlib_plot_assembly_blocks(
    rectangles,
    guides,
    tops,
    ax=None,
    fig_height="auto",
    fig_width=8,
    legend=False,
    textprops=None,
    record=None,
    legend_offset=-0.05
):
    """Plot the assembly block rectangles using matplotlib."""

    if record is not None:
        fig, record_ax, blocks_ax = plt.subplots(
            2, 1, figsize=(7, 7), gridspec_kw={"height_ratios": [4, 1]}
        )
        grecord = GraphicRecord.from_biopython_record(
            record, fun_color=lambda a: "#aabbff"
        )
        grecord.plot(record_ax)
        res = _matplotlib_plot_assembly_blocks(
            rectangles,
            guides,
            tops,
            ax=blocks_ax,
            fig_height="auto",
            fig_width=8,
            legend=False,
            textprops=None,
            record=None,
        )
        record_ax.set_xlim(blocks_ax.get_xlim())
        return res

    L = max([r["right"] for r in rectangles])

    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"], size=12, family="sans-serif"
        )

    if ax is None:
        if fig_height == "auto":
            fig_height = 8
        fig, ax = plt.subplots(1, figsize=(8, 8))
    ax.set_xlim((-0.05 * L, 1.05 * L))
    ax.set_ylim((min(tops) - 1, 1.5))

    for g in guides:
        patch = mpatches.Rectangle(
            (g["left"], g["bottom"]),
            g["right"] - g["left"],
            g["top"] - g["bottom"],
            color=g["color"],
            ec="none",
        )
        ax.add_patch(patch)

    seen_sources = set([])
    legend_handles = []
    sorted_rectangles = sorted(
        rectangles, key=lambda r: (-r["bottom"], r["left"])
    )
    for g in sorted_rectangles:
        if g["source"] not in seen_sources:
            legend_patch = mpatches.Patch(
                facecolor=g["color"],
                label=g["source"],
                edgecolor="k",
                linewidth=1.0,
            )
            legend_handles.append(legend_patch)
            seen_sources.add(g["source"])
            plt.legend(handles=[legend_patch])
        # linewidth = g["line_width"]
        width = g["right"] - g["left"]
        line_width = 1.0 if (1.0 * width / L) > 0.002 else 0
        patch = mpatches.Rectangle(
            (g["left"], g["bottom"]),
            width,
            g["top"] - g["bottom"],
            facecolor=g["color"],
            edgecolor="k",
            linewidth=line_width,
        )
        ax.add_patch(patch)
    ax.axis("off")

    if legend:
        if ax.legend_ is not None:
            ax.legend_.remove()
        legend = ax.legend(
            handles=legend_handles,
            frameon=False,
            ncol=2,
            loc=2,
            bbox_to_anchor=(0.0, legend_offset),
        )
        ltext = legend.get_texts()
        plt.setp(ltext, fontproperties=textprops)
        ax.figure.subplots_adjust(hspace=0.0)

    return ax, legend_handles


from ..DnaSupplier import DnaSupplier
from ...DnaQuote import DnaQuote


class PcrLinearizationStation(DnaSupplier):
    """Station which will linearize another station's (circular) output via PCR.
    
    Parameters
    ----------
    """

    report_fa_symbol = u""
    report_fa_symbol_plain = "exchange"
    report_color = "#eefefe"
    operation_type = "PCR"

    def __init__(
        self,
        name,
        supplier,
        primers_supplier,
        homology_selector,
        sequence_constraints=(),
        extra_time=0,
        extra_cost=0,
        memoize=False,
    ):
        self.name = name
        self.supplier = supplier
        self.primers_supplier = primers_supplier
        self.homology_selector = homology_selector
        self.sequence_constraints = sequence_constraints
        self.extra_time = extra_time
        self.extra_cost = extra_cost
        self.memoize = memoize

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False,
    ):
        """Return a price-optimal DnaQuote for the given sequence.

        It will find a possible hit in the blast database, find the primers to
        order for the PCR, compute the overall price and lead time, and return
        a quote.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quote

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """
        if max_lead_time is not None:
            suppliers_max_lead_time = max_lead_time - self.extra_time
        else:
            suppliers_max_lead_time = None
        supplier_quote = self.supplier.get_quote(
            sequence, max_lead_time=suppliers_max_lead_time
        )
        if not supplier_quote.accepted:
            return DnaQuote(
                self, sequence, accepted=False, message="Refused by supplier"
            )

        left_location = self.homology_selector.compute_segment_location(
            sequence=sequence, index=0
        )
        right_location = self.homology_selector.compute_segment_location(
            sequence=sequence, index=len(sequence)
        )
        if (left_location is None) or (right_location is None):
            return DnaQuote(
                self,
                sequence,
                accepted=False,
                message="Could not design a suitable primer",
            )
        _, left_end = left_location
        right_start, _ = right_location

        left_primer = sequence[:left_end]
        right_primer = sequence[right_start:]

        primers_quotes = [
            self.primers_supplier.get_quote(
                primer, max_lead_time=suppliers_max_lead_time
            )
            for primer in [left_primer, right_primer]
        ]
        if not all(quote.accepted for quote in primers_quotes):
            return DnaQuote(
                self,
                sequence,
                accepted=False,
                message="Primers refused by primers supplier",
            )
        all_quotes = [supplier_quote] + primers_quotes
        if max_lead_time is not None:
            overall_lead_time = (
                max(quote.lead_time for quote in all_quotes) + self.extra_time
            )
        else:
            overall_lead_time = None
        total_price = (
            sum(quote.price for quote in all_quotes) + self.extra_cost
        )

        if with_assembly_plan:
            assembly_plan = {
                (0, len(sequence)): all_quotes[0],
                (0, int(left_end)): all_quotes[1],
                (int(right_start), len(sequence)): all_quotes[2],
            }
        else:
            assembly_plan = None

        return DnaQuote(
            self,
            sequence,
            accepted=True,
            lead_time=overall_lead_time,
            price=total_price,
            assembly_plan=assembly_plan,
            metadata={"location": (0, len(sequence)),},
        )

    def additional_dict_description(self):
        return {
            "primers DNA source": self.primers_supplier.name,
            "primers melting temp.": self.primer_melting_temperature,
        }



from ..DnaSupplier import DnaSupplier


class SequenceAdapter(DnaSupplier):
    """Transparent source adding sequence flanks between suppliers and clients.
    """

    class_description = "Adapts a sequence between a supplier and its clients"
    operation_type = "sequence extension"
    report_fa_symbol = u"\uf07e"
    report_fa_symbol_plain = "arrows-h"
    report_color = "#ffeeee"

    def __init__(
        self,
        name,
        supplier,
        left_addition="",
        right_addition="",
        memoize=False,
    ):
        self.name = name
        self.supplier = supplier
        self.left_addition = left_addition
        self.right_addition = right_addition
        self.sequence_constraints = ()
        self.memoize = memoize
        if hasattr(supplier, "min_basepair_price"):
            self.min_basepair_price = supplier.min_basepair_price

    def __repr__(self):
        return self.name

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False,
    ):
        extended_sequence = self.left_addition + sequence + self.right_addition
        return self.supplier.get_quote(
            extended_sequence,
            max_lead_time=max_lead_time,
            with_assembly_plan=with_assembly_plan,
        )

    @staticmethod
    def from_dict(data):
        return SequenceAdapter(
            name=data["name"],
            supplier=data["supplier"],
            left_addition=data["left_addition"],
            right_addition=data["right_addition"],
        )

    def suggest_cuts(self, sequence):
        return self.supplier.suggest_cuts(sequence)



from ...DnaQuote import DnaQuote
from ...SegmentSelector import TmSegmentSelector, FixedSizeSegmentSelector
from ...biotools import (
    reverse_complement,
    largest_common_substring,
    blast_sequence,
    perfect_match_locations_in_hsp,
)
from ...tools import functions_list_to_string
from ..DnaSupplier import DnaSupplier
from ..builtin_constraints import SequenceLengthConstraint
from .DnaSuppliersComparator import DnaSuppliersComparator


class PcrExtractionStation(DnaSupplier):
    """Class to represent databases of constructs which can be (in part) reused

    A blast database contains the sequences of all available constructs.
    Given a sequence, the PcrExtractionStation finds whether it is possible to order
    two primers to extract this sequence from the constructs in the BLAST
    database.

    Parameters
    ----------

    name
      Name of the PCR station (e.g. "Lab constructs PCR station")

    primers_supplier
      DnaSupplier providing the primers (will typically be an CommercialDnaOffer)

    blast_database

    sequences
      A dictionary {seq_name: sequence_in_atgc}

    pcr_homology_length

    max_overhang_length

    extra_cost

    extra_time

    max_amplicon_length

    blast_word_size
      BLAST parameter, the larger number will lead to faster BLASTing but with
      a risk of missing short sequences of length shorter than this value. 

    memoize

    sequence_constraints

    """

    class_description = "PCR-extraction station"
    operation_type = "PCR"
    report_fa_symbol = u""
    report_fa_symbol_plain = "recycle"
    report_color = "#eeffee"
    dna_banks = {}

    def __init__(
        self,
        name,
        primers_supplier,
        homology_selector,
        blast_database=None,
        sequences=None,
        pcr_homology_length=25,
        max_overhang_length=40,
        extra_cost=0,
        extra_time=0,
        max_amplicon_length=None,
        blast_word_size=50,
        memoize=False,
        sequence_constraints=(),
    ):
        self.name = name
        self.homology_selector = homology_selector
        self.blast_database = blast_database
        self.set_suppliers(primers_supplier)
        self.pcr_homology_length = pcr_homology_length
        self.max_overhang_length = max_overhang_length
        self.extra_time = extra_time
        self.extra_cost = extra_cost
        self.max_amplicon_length = max_amplicon_length
        self.blast_word_size = blast_word_size
        self.sequence_constraints = list(sequence_constraints)
        if max_amplicon_length is not None:
            c = SequenceLengthConstraint(max_length=max_amplicon_length)
            self.sequence_constraints = [c] + self.sequence_constraints
            self.min_basepair_price = (
                2 * 20 * self.primers_supplier.min_basepair_price
            ) / self.max_amplicon_length
        self.memoize = memoize
        self.memoize_dict = {}
        self.sequences = sequences

    def _get_hits(self, sequence):
        """Return the hits of the given sequence against the blast database
        in format [(part_hit, (start, end), subsequence)
        
        
        
        """
        if self.sequences is not None:
            result = []
            for dna_name, seq in self.sequences.items():
                match_coords = largest_common_substring(
                    sequence, seq, self.max_overhang_length
                )
                # if match_coords
                if match_coords:
                    result.append((dna_name, match_coords, None))
            return result
        else:
            record = blast_sequence(
                sequence,
                self.blast_database,
                perc_identity=98,
                use_megablast=True,
                word_size=self.blast_word_size,
            )

            return [
                (
                    al.hit_id + "_h%03d" % i,
                    (hit.query_start, hit.query_end),
                    hit.sbjct,
                )
                for al in record.alignments
                for i, hit in enumerate(al.hsps)
            ]

    def blast_sequence(self, sequence, cutoff=40):
        if hasattr(sequence, 'seq'):
            sequence = str(sequence.seq)
        record = blast_sequence(
            sequence,
            self.blast_database,
            perc_identity=98,
            use_megablast=True,
            word_size=self.blast_word_size,
        )
        return [
            (
                al.hit_id + "_h%03d_%02d" % (i, j),
                (start, end),
                sequence[start:end],
            )
            for al in record.alignments
            for i, hit in enumerate(al.hsps)
            for j, (start, end) in enumerate(
                perfect_match_locations_in_hsp(hit, span_cutoff=cutoff)
            )
        ]

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False,
    ):
        """Return a price-optimal DnaQuote for the given sequence.

        It will find a possible hit in the blast database, find the primers to
        order for the PCR, compute the overall price and lead time, and return
        a quote.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quote

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """
        hits = self._get_hits(sequence)

        # For each hit in the database, see if there is a
        for subject, (hit_start, hit_end), _ in hits:
            if min(hit_start, hit_end) < 0:
                continue
            largest_overhang = max(hit_start, len(sequence) - hit_end)

            if largest_overhang > self.max_overhang_length:
                continue
            for i in range(
                min(len(sequence), self.max_overhang_length) - hit_start
            ):
                subseq = sequence[hit_start + i :]
                # print (hit_start, i, subseq, hit_end, hit_start)
                left_location = self.homology_selector.compute_segment_location(
                    subseq, 0
                )
                if left_location is not None:
                    l_start, l_end = left_location
                    primer_left = sequence[: l_end + hit_start + i]
                    break
            else:
                continue
            right_padding = len(sequence) - hit_end
            for i in range(self.max_overhang_length - right_padding):
                subseq = sequence[: hit_end - i]
                right_location = self.homology_selector.compute_segment_location(
                    subseq, hit_end - i
                )
                if right_location is not None:
                    r_start, r_end = right_location
                    primer_right = reverse_complement(sequence[r_start:])
                    break
            else:
                continue
            # primer_l_end = hit_start + self.pcr_homology_length
            # primer_left = sequence[:primer_l_end]
            # primer_r_end = hit_end - self.pcr_homology_length
            # primer_right = reverse_complement(sequence[primer_r_end:])

            primer_max_lead_time = (
                None
                if max_lead_time is None
                else max_lead_time - self.extra_time
            )
            quotes = [
                self.primers_supplier.get_quote(
                    primer, max_lead_time=primer_max_lead_time
                )
                for primer in [primer_left, primer_right]
            ]
            if not all(quote.accepted for quote in quotes):
                continue  # primers inorderable

            if max_lead_time is not None:
                overall_lead_time = (
                    max(quote.lead_time for quote in quotes) + self.extra_time
                )
            else:
                overall_lead_time = None
            total_price = (
                sum(quote.price for quote in quotes) + self.extra_cost
            )

            if with_assembly_plan:
                assembly_plan = {
                    (0, int(l_end)): quotes[0],
                    (int(r_start), len(sequence)): quotes[1],
                }
            else:
                assembly_plan = None

            return DnaQuote(
                self,
                sequence,
                accepted=True,
                lead_time=overall_lead_time,
                price=total_price,
                assembly_plan=assembly_plan,
                message="From %s" % subject,
                metadata={
                    "subject": subject,
                    "location": (hit_start, hit_end),
                },
            )
        if len(hits):
            message = (
                "Some matches found but could not find suitable primer "
                "design."
            )
        else:
            message = "No BLAST hit found"
        return DnaQuote(self, sequence, accepted=False, message=message)

    def suggest_cuts(self, sequence):
        if self.sequences is None:
            return []
        suggested_cuts = []
        for name, subseq in self.sequences.items():
            index = sequence.find(subseq)
            if index >= 0:
                suggested_cuts.extend([index, index + len(subseq)])
        return sorted(list(set(suggested_cuts)))

    def prepare_on_sequence(self, sequence):
        """Pre-compute the BLAST of the current sequence against the database.

        Once a pre-blast has been performed, this PcrExtractionStation becomes
        specialized on that sequence and its subsequences, do not feed it with
        another different sequence. Do `self.sequences=None` to reinitialize
        and de-specialize this PcrExtractionStation.

        Examples
        --------

        >>> pcr_station = PcrExtractionStation("some_blast_database")
        >>> top_station = # some assembly station depending on pcr_station
        >>> pcr_station.prepare_on_sequence(my_sequence)
        >>> top_station.get_quote(my_sequence)
        >>> pcr_station.sequences=None # de-specializes the pcr station.
        """
        if self.blast_database is None:
            return
        hits = self.blast_sequence(sequence)
        self.sequences = None  # destroy current pre-blast (used by _get_hits)
        self.sequences = {subject: seq for subject, (start, end), seq in hits}

    def additional_dict_description(self):
        constraints = functions_list_to_string(self.sequence_constraints)
        return {
            "BLAST database": self.blast_database,
            "primers dna source": self.primers_supplier.name,
            "pcr homology length": self.pcr_homology_length,
            "max overhang length": self.max_overhang_length,
            "extra time": self.extra_time,
            "extra cost": self.extra_cost,
            "max amplicon length": self.max_amplicon_length,
            "BLAST word size": self.blast_word_size,
            "sequence constraints": constraints,
        }

    @classmethod
    def from_dict(cls, data):
        if "dna_bank" in data:
            blast_database = cls.dna_banks[data["dna_bank"]]
        else:
            blast_database = data["blast_database"]

        if data["homology_type"] == "tm":
            min_oh_size, max_oh_size = data["homology_size_range"]
            min_tm, max_tm = data["tm_range"]
            homology_selector = TmSegmentSelector(
                min_size=min_oh_size,
                max_size=max_oh_size,
                min_tm=min_tm,
                max_tm=max_tm,
            )
        else:
            homology_selector = FixedSizeSegmentSelector(
                segment_size=data["homology_size"]
            )
        return PcrExtractionStation(
            name=data["name"],
            homology_selector=homology_selector,
            primers_supplier=data["suppliers"],
            max_overhang_length=data["max_overhang_length"],
            max_amplicon_length=data["max_amplicon_length"],
            extra_cost=data["cost"],
            blast_database=blast_database,
        )

    def set_suppliers(self, suppliers):
        if hasattr(suppliers, "__iter__"):
            if len(suppliers) > 1:
                self.primers_supplier = DnaSuppliersComparator(
                    name=self.name + " comparator", suppliers=suppliers
                )
            else:
                self.primers_supplier = suppliers[0]
        else:
            self.primers_supplier = suppliers



import os
import flametree
import pandas
from ..config import SETTINGS

class FolderReportMixin:

    def write_full_report(self, target="@memory"):
        """Generate an extensive, multifile report as either a folder or a zip."""

        def write_ax_as_pdf(ax, target):
            ax.figure.savefig(target, format="pdf", bbox_inches="tight")

        root = flametree.file_tree(target, replace=True)
        self.write_pdf_report(root._file("assembly_report.pdf").open("wb"))
        self.write_sequences_spreadsheet(root._file('sequences.csv').open("w"))
        root._file("README.txt").write(_get_folder_readme_content())
        self.write_figures_folder(root._dir("figures"))
        self.write_all_sequence_records(root._dir("genbank"))
        return root._close()
    




    def write_sequences_spreadsheet(quote, target):

        quotes_list = quote.to_steps_list()
        sequences_dataframe = pandas.DataFrame.from_records(
            [
                {
                    "ID": _quote.id,
                    "Type": _quote.get("operation_type", None),
                    "Source": _quote.source,
                    "SequenceLength": len(_quote.sequence),
                    "Sequence": _quote.sequence
                }
                for _quote in quotes_list
            ],
            columns=["ID", "Type", "Source", "SequenceLength", "Sequence"]
        ).sort_values(by=["Type", "Source", "ID"])

        sequences_dataframe.to_csv(target, sep=";", index=False)

def _get_folder_readme_content():
    """General function to get the content of the FOLDER_README.txt template"""
    readme_path = os.path.join(SETTINGS["template_path"], "FOLDER_README.txt")
    with open(readme_path, "r") as f:
        readme_content = f.read()
    return readme_content



"""This class implements a very thin layer to store metadata in a string"""

class SequenceString(str):
    def __new__(self, value, metadata=None):
        return str.__new__(self, value)

    def __init__(self, value, metadata=None):
        self.metadata = metadata or {}
    
    @staticmethod
    def from_record(record, topology="default_to_linear"):
        if topology.startswith("default_to_"):
            default_topology = topology.split("_")[-1]
            topology = record.annotations.get("topology", default_topology)
        return SequenceString(str(record.seq), metadata={"topology": topology})


# Attempt import of optional module DNA Chisel
try:
    from dnachisel import Specification, DnaOptimizationProblem

    DNACHISEL_AVAILABLE = True
except:
    Specification = type("BLANK")  # fake type that matches nothing
    DNACHISEL_AVAILABLE = False

class ConstraintsMixin:

    def verify_constraints(self, sequence):
        """Return True iff `sequence` passes all `self.sequence_constraints`

        Will automatically process DNA-Chisel constraints that would be in
        `self.sequence_constraints`

        """
        constraints = self.sequence_constraints
        if not hasattr(self, "dnachisel_constraints"):
            self.dnachisel_constraints = [
                constraint
                for constraint in self.sequence_constraints
                if isinstance(constraint, Specification)
            ]

        if self.dnachisel_constraints != []:
            if not DNACHISEL_AVAILABLE:
                raise ImportError(
                    "Spotted DNA Chisel constraints, while "
                    "DNA Chisel is not installed."
                )
            # We provide an empty mutation space so it won't be recomputed
            # (which would take time and is useless here!)
            problem = DnaOptimizationProblem(
                sequence, self.dnachisel_constraints, mutation_space=[]
            )
            constraints = [
                constraint
                for constraint in constraints
                if not isinstance(constraint, Specification)
            ] + [lambda seq: problem.all_constraints_pass()]

        return all(constraint(sequence) for constraint in constraints)


"""Classes to represent assembly strategies in DnaWeaver."""


from collections import defaultdict
from .ExportsMixin import ExportsMixin
from .PostProcessingMixin import PostProcessingMixin




class DnaQuote(ExportsMixin, PostProcessingMixin):
    """Class to represent the Quote returned by a DNA source in response to
    a quote request for a given DNA sequence.

    Parameters
    -----------

    source
      The DnaSupplier which issued the quote

    sequence

    price

    accepted

    lead_time

    assembly_plan
      A dict of the form { (start1, end1): op1, (start2, end2): op2, ...}
      where the `(start, end)` represent sub-segments of the final sequence and
      `op1, op2` represent the AssemblyOperations that create the corresponding
      fragments

     metadata

     message

     id
       A string identifying uniquely the assembly operation


     deadline
       Deadline for the operation (see the `propagate_deadline` method for this
       class) i.e. in how many time units it should be finished. This is for
       drawing assembly timelines.

    """

    def __init__(
        self,
        source,
        sequence,
        price=None,
        accepted=True,
        lead_time=None,
        assembly_plan=None,
        metadata={},
        message="",
        id=None,
        deadline=None,
    ):
        self.source = source
        self.accepted = accepted
        self.price = price
        self.lead_time = lead_time
        self.sequence = sequence
        self.message = message
        self.assembly_plan = assembly_plan
        self.full_assembly_plan_computed = False
        self.metadata = metadata
        self.id = id
        self.deadline = None

    def __repr__(self):
        if self.accepted:
            infos = "price %.02f" % self.price
            if self.lead_time is not None:
                infos = infos + " - lead_time %0.1f" % self.lead_time
        else:
            infos = "refused: %s" % self.message
        result = "From %s - %s" % (self.source, infos)
        if len(self.message):
            result += " - " + self.message
        return result

    @property
    def step_duration(self):
        """Duration of the final step of the assembly in the quote.

        Inferred by substracting the lead time of children operations to the
        lead time of this operation.
        """
        children_lead_time = self.children_overall_lead_time()
        if children_lead_time is None:
            return self.lead_time
        else:
            return self.lead_time - self.children_overall_lead_time()

    def compute_assembly_levels(self):
        """Return edges and levels for drawing the assembly graph nicely.

        """

        levels = defaultdict(lambda: [])
        edges = []

        def rec(subtree, depth=0):

            levels[depth].append(subtree)

            if subtree.assembly_plan is not None:
                for other in subtree.assembly_plan.values():
                    edges.append((other, subtree))
                    rec(other, depth + 1)

        rec(self)
        levels = [levels[i] for i in sorted(levels.keys())][::-1]
        levels = [sorted(level, key=lambda e: e.id)[::-1] for level in levels]
        return edges, levels

    def assembly_step_summary(self):
        """Return a print-friendly, simple string of the ordering plan."""
        plan = "\n  ".join(
            "%s-%s: %s" % (start, end, str(quote))
            for (start, end), quote in sorted(self.assembly_plan.items())
        )
        final_txt = "Ordering plan:\n  %s\nPrice:%.02f" % (plan, self.price)
        if self.lead_time is not None:
            final_txt = final_txt + ", total lead_time:%.1f" % self.lead_time
        return final_txt

    def children_total_price(self):
        """Return the total price of all sub-operations (apart from current)"""
        # print ([quote.accepted for quote in self.assembly_plan.values()])
        return sum(quote.price for quote in self.assembly_plan.values())

    def children_overall_lead_time(self):
        """Return the max lead time of all sub-operation (current one not
        included)"""
        if self.assembly_plan is None:
            return None
        lead_times = [quote.lead_time for quote in self.assembly_plan.values()]
        if len(lead_times) == 0:
            return None
        elif any(lead_time is None for lead_time in lead_times):
            return None
        else:
            return max(lead_times)

    def cuts_indices(self):
        """Return the locations of the `cuts` where the sequence is assembled.
        """
        return sorted([segment[1] for segment in self.assembly_plan])[:-1]



from copy import deepcopy
from io import StringIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import DNAAlphabet
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation


class GenbankExportMixin:
    def to_record(self, record=None, record_id=None):
        """Return a Biopython seqrecord of the quote.

        >>> record = to_record(solution)
        >>> # Let's plot with DnaVu:
        >>> from dnavu import create_record_plot
        >>> from bokeh.io import output_file, show
        >>> output_file("view.html")
        >>> plot = create_record_plot(record)
        >>> show(plot)
        """
        if record_id is None:
            record_id = self.id
        if record is None:
            record = SeqRecord(Seq(self.sequence, DNAAlphabet()), id=record_id)
        else:
            record = deepcopy(record)

        if self.plan is not None:
            features = [
                SeqFeature(
                    FeatureLocation(q.segment_start, q.segment_end, 1),
                    type="misc_feature",
                    qualifiers={
                        "label": "%s - From %s" % (q.id, q.source),
                        "name": q.id,
                        "source": q.source,
                        "price": q.price,
                        "lead_time": q.lead_time,
                    },
                )
                for q in self.plan
            ]
            record.features = features + record.features
        return record

    def write_genbank(
        self, filename=None, filehandle=None, record=None, record_id=None
    ):
        record = self.to_record(record=record, record_id=record_id)
        if filehandle is None:
            with open(filename, "w+") as f:
                SeqIO.write(record, f, "genbank")
        else:
            SeqIO.write(record, filehandle, "genbank")

    def write_all_sequence_records(self, target):
        for step in self.to_steps_list():
            record = self.plan_step_to_record(step)
            path = target._file(step.id + ".gb").open("w")
            SeqIO.write(record, path, "genbank")

    @staticmethod
    def plan_step_to_record(plan_step, record=None, record_id=None):
        """Return a Biopython seqrecord of the quote.
        >>> record = to_SeqRecord(solution)
        >>> # Let's plot with DnaVu:
        >>> from dnavu import create_record_plot
        >>> from bokeh.io import output_file, show
        >>> output_file("view.html")
        >>> plot = create_record_plot(record)
        >>> show(plot)
        """
        if record_id is None:
            record_id = plan_step.id
        if record is None:
            record = SeqRecord(
                Seq(plan_step.sequence, DNAAlphabet()), id=record_id
            )
        else:
            record = deepcopy(record)

        if plan_step.assembly_plan is not None:
            features = [
                SeqFeature(
                    FeatureLocation(q.segment_start, q.segment_end, 1),
                    type="misc_feature",
                    qualifiers={
                        "label": "%s - From %s" % (q.id, q.source),
                        "name": q.id,
                        "source": q.source,
                        "price": q.price,
                        "lead_time": q.lead_time,
                    },
                )
                for q in plan_step.assembly_plan
            ]
            record.features = features + record.features
        return record



import tempfile
import time
import os
import subprocess
from Bio.Blast import NCBIXML
import numpy as np
from .sequence_operations import sequence_to_atgc

def blast_sequence(
    sequence,
    blast_db=None,
    subject=None,
    word_size=4,
    perc_identity=80,
    num_alignments=1000,
    num_threads=3,
    use_megablast=True,
    ungapped=True,
):
    """Return a Biopython BLAST record of the given sequence BLASTed
    against the provided database.

    Parameters
    ----------

    sequence
      An ATGC sequence

    blast_db
      Path to a BLAST database

    subject
      Either a path to a fasta (.fa) file or an ATGC string. Subject to blast
      against.

    word_size
      Word size to use in the blast

    perc_identity
      Minimal percentage of identical nucleotides in a match for it to be kept

    num_alignments
      Number of alignments to keep

    num_threads
      Number of threads for the BLAST

    use_megablast
      Whether to use Megablast.

    ungapped
      No-gaps matches only ?

    Examples
    --------

    >>> blast_record = blast_sequence("ATTGTGCGTGTGTGCGT", "blastdb/ecoli")
    >>> for alignment in blast_record.alignments:
    >>>     for hit in alignment.hsps:
    >>>         print (hit.identities)
    """

    xml_file, xml_name = tempfile.mkstemp(".xml")
    fasta_file, fasta_name = tempfile.mkstemp(".fa")
    sequence = sequence_to_atgc(sequence)
    with open(fasta_name, "w+") as f:
        f.write(">seq\n" + sequence)

    if subject is not None:
        close_subject = True
        if not subject.endswith(".fa"):
            remove_subject = True
            _subject_file, fasta_subject_name = tempfile.mkstemp(".fa")
            with open(fasta_subject_name, "w+") as f:
                f.write(">subject\n" + subject)
            subject = fasta_subject_name
        else:
            remove_subject = False
    else:
        close_subject = False
    p = subprocess.Popen(
        [
            "blastn",
            "-out",
            xml_name,
            "-outfmt",
            "5",
            "-num_alignments",
            str(num_alignments),
            "-query",
            fasta_name,
        ]
        + (
            ["-db", blast_db]
            if blast_db is not None
            else ["-subject", subject]
        )
        + (["-ungapped"] if ungapped else [])
        + (["-task", "megablast"] if use_megablast else [])
        + [
            "-word_size",
            str(word_size),
            "-num_threads",
            str(num_threads),
            "-dust",
            "no",
            "-evalue",
            "0.01",
            "-perc_identity",
            str(perc_identity),
        ],
        close_fds=True,
        stderr=subprocess.PIPE,
    )
    res, _blast_err = p.communicate()
    p.wait()
    error = None
    for i in range(3):
        try:
            with open(xml_name, "r") as f:
                res = list(NCBIXML.parse(f))
                os.fdopen(xml_file, "w").close()
                os.fdopen(fasta_file, "w").close()
                os.remove(xml_name)
                os.remove(fasta_name)

                if close_subject:
                    open(subject, "w").close()
                    if remove_subject:
                        os.remove(subject)
                if len(res) == 1:
                    return res[0]
                else:
                    return res
            break
        except ValueError as err:
            error = err
            time.sleep(0.1)
    else:
        raise ValueError("Problem reading the blast record: " + str(error))

def make_blast_db(fasta_input, target):
    proc = subprocess.Popen(
        ["makeblastdb", "-in", fasta_input, "-dbtype", "nucl", "-out", target]
    )
    proc.wait()


def perfect_match_locations_in_hsp(hsp, span_cutoff=10):
    """Return the locations of perfect matches in a BLAST HSP.
    
    Only locations with a span above span_cutoff are kept.
    """
    if hsp.align_length < span_cutoff:
        return []
    arr = np.frombuffer(hsp.match.encode(), dtype="uint8")
    indices = [0] + list((arr != 124).nonzero()[0]) + [len(arr)]
    return [
        (start + hsp.query_start, end + hsp.query_start)
        for start, end in zip(indices, indices[1:])
        if end - start >= span_cutoff
    ]


def largest_common_substring(query, target, max_overhang):
    """Return the largest common substring between `query` and `target`.

    Find the longest substring of query that is contained in target.

    If the common substring is too much smaller than `query` False is returned,
    else the location `(start, end)` of the substring in `target` is returned.

    Parameters:
    -----------

    query (str)
      The sequence to be found in target (minus some overhangs possibly)

    target (str)
      The sequence in which to find `query`

    max_overhang
      Maximal size allowed for the flanking regions of `query` that would
      not be contained in `target`.
    
    Examples
    --------

    >>> seqA = '-----oooooooo'
    >>> seqB = 'oooooo-----tttt'
    >>> largest_common_substring(seqA, seqA, 80) # == (0, 12)
    >>> largest_common_substring(seqA, seqB, 80) # == (5, 11)

    Notes:
    ------

    This is intended for finding whether `query` can be extracted from `target`
    using PCR. See the PcrExtractionStation implementation in DnaSupplier.py.

    """
    # The trick here is to start with the central region of "query".
    # This region is initially as small as max_overhang allows, and it is
    # progressively expanded on the sides
    max_overhang = min(max_overhang, int(len(query) / 2))
    start, end = max_overhang, len(query) - max_overhang
    if query[start:end] not in target:
        return False
    while (start >= 0) and (query[start:end] in target):
        start -= 1
    start += 1
    while (end < len(query)) and (query[start:end] in target):
        end += 1
    end -= 1
    return start, end


"""Plotting functions for the results of DNA Weaver"""

from io import StringIO, BytesIO
from base64 import b64encode



def matplotlib_figure_to_file_string(fig, format="svg", **kwargs):
    """Return a string of the figure in the requested format."""

    if (format == "pdf"):
        output = BytesIO()
    else:
        output = StringIO()

    fig.savefig(output, format=format, **kwargs)
    return output.getvalue()


def matplotlib_figure_to_svg_base64_data(fig, **kwargs):
    """Return a string of the form 'data:image/svg+xml;base64,XXX' where XXX
    is the base64-encoded svg version of the figure."""
    svg_txt = matplotlib_figure_to_file_string(fig, format="svg", **kwargs)
    svg_txt = "\n".join(svg_txt.split("\n")[4:])
    svg_txt = "".join(svg_txt.split("\n"))
    try:
        return "data:image/svg+xml;base64," + b64encode(svg_txt)
    except:
        content = b64encode(svg_txt.encode("ascii"))
        result = (b"data:image/svg+xml;base64," + content).decode("utf-8")
        return str(result)



from collections import defaultdict
from copy import deepcopy
import matplotlib.font_manager as fm
import numpy as np
from ...config import SETTINGS
from .plot_tree_graph import plot_tree_graph

class AssemblyGraphMixin:
    def plot_assembly_graph(
        self, ax=None, margin=None, textprops=None, scale=1.0
    ):
        """Plot the complete assembly graph
        
        Returns
        -------
        elements_positions, ax
        Dictionary of elements positions, matplotlib ax.
        """

        nodes_dict = {}
        levels = defaultdict(lambda *a: [])
        edges = []
        tree = deepcopy(self.plan)

        def rec(node, depth=0):
            if node.get("_visited", False):
                return
            nodes_dict[node.id] = node
            node["_visited"] = True
            assembly_plan = node.pop("assembly_plan")
            levels[depth].append(node.id)
            for other in assembly_plan:
                edges.append([other.id, node.id])
                rec(other, depth + 1)

        rec(tree)
        levels = [levels[i] for i in range(max(levels) + 1)][::-1]

        fontawesome = fm.FontProperties(
            fname=SETTINGS["fontawesome-ttf-path"],
            size=13 * scale,
            family="sans-serif",
        )
        if textprops is None:
            textprops = fm.FontProperties(
                fname=SETTINGS["OpenSans-ttf-path"],
                size=12 * scale,
                family="sans-serif",
            )

        def draw_node(x, y, node_id, ax):
            node = nodes_dict[node_id]
            icon = self.sources[node.source]._report_fa_symbol
            ax.text(
                x,
                y,
                node_id,
                horizontalalignment="left",
                verticalalignment="center",
                fontproperties=textprops,
            )
            ax.text(
                x - 0.01 * np.sqrt(scale),
                y,
                icon,
                horizontalalignment="right",
                verticalalignment="center",
                fontproperties=fontawesome,
            )

        all_elements = sorted(sum(levels, []))
        ypos = {
            el: 1.0 * (i + 1) / (len(all_elements) + 2)
            for i, el in enumerate(all_elements)
        }
        for el in all_elements:
            children = [e2 for (e2, e1) in edges if e1 == el]
            if children != []:
                ypos[el] = 1.0 * sum(ypos[e] for e in children) / len(children)

        xpos = {
            el: 1.0 * (1 + x) / (len(levels) + 1)
            for x, elements in enumerate(levels)
            for el in elements
        }

        elements_positions = {el: (xpos[el], ypos[el]) for el in all_elements}
        return plot_tree_graph(
            levels,
            edges,
            draw_node,
            elements_positions=elements_positions,
            ax=ax,
            edge_left_space=0.06,
            edge_right_space=0.03,
            margin=margin,
            height_factor=0.40,
            width_factor=5.5,
            scale=scale,
        )



from .SequenceAdapter import SequenceAdapter
from .CommercialDnaOffer import CommercialDnaOffer
from .DnaAssemblyStation import DnaAssemblyStation
from .DnaSuppliersComparator import DnaSuppliersComparator
from .PartsLibrary import PartsLibrary, GoldenGatePartsLibrary
from .PcrExtractionStation import PcrExtractionStation
from .PcrLinearizationStation import PcrLinearizationStation

__all__ = [
    "SequenceAdapter",
    "CommercialDnaOffer",
    "DnaAssemblyStation",
    "DnaSuppliersComparator",
    "PartsLibrary, GoldenGatePartsLibrary",
    "PcrExtractionStation",
    "PcrLinearizationStation"
]


from .FixedSizeSegmentSelector import FixedSizeSegmentSelector
from .SegmentSelector import SegmentSelector
from .TmSegmentSelector import TmSegmentSelector

__all__ = [
    "FixedSizeSegmentSelector",
    "SegmentSelector",
    "TmSegmentSelector"
]


import itertools as itt
import os
import tempfile
from ..biotools import blast_sequence

class PostProcessingMixin:


    def compute_full_assembly_plan(self, id_prefix="S", id_digits=5):
        """ """

        counter = itt.count()

        def rec(quote):
            if not quote.accepted:
                return quote

            if any(
                [
                    hasattr(quote.source, attr)
                    for attr in ["supplier", "primers_supplier"]
                ]
            ):
                if quote.assembly_plan is None:
                    quote = quote.source.get_quote(
                        quote.sequence,
                        max_lead_time=quote.lead_time,
                        with_assembly_plan=True,
                    )
                segments = {
                    segment: rec(subquote)
                    for segment, subquote in sorted(
                        quote.assembly_plan.items(), key=lambda item: item[0]
                    )
                }
                quote.assembly_plan = segments
            if id_prefix:
                index = next(counter)
                quote.id = "{id_prefix}_{index:0{id_digits}}".format(
                    id_prefix=id_prefix, index=index, id_digits=id_digits
                )
            return quote

        rec(self)
        

        if id_prefix:
            index = next(counter)
            self.id = "{id_prefix}_{index:0{id_digits}}".format(
                id_prefix=id_prefix, index=index, id_digits=id_digits
            )
        self.full_assembly_plan_computed = True

    def compute_fragments_final_locations(self):
        """Compute the exact final location of the fragments in the final
        sequence.
        """
        if not self.full_assembly_plan_computed:
            self.compute_full_assembly_plan()
        quotes = self.tree_as_list()
        quotes_dict = {quote.id: quote for quote in quotes}
        _, temp_fasta = tempfile.mkstemp(suffix=".fa")
        with open(temp_fasta, "w+") as f:
            for quote in quotes:
                f.write(">%s\n%s\n" % (quote.id, quote.sequence))
        results = blast_sequence(
            self.sequence, subject=temp_fasta, word_size=10, perc_identity=100
        )

        if isinstance(results, list):
            alignments = sum([rec.alignments for rec in results], [])
        else:
            alignments = results.alignments

        for al in alignments:
            hit = max(al.hsps, key=lambda hit: hit.align_length)
            final_location = sorted((hit.query_start, hit.query_end))
            matching_segment = sorted((hit.sbjct_start, hit.sbjct_end))
            quotes_dict[al.hit_def].final_location = final_location
            quotes_dict[al.hit_def].matching_segment = matching_segment
        os.remove(temp_fasta)
    

    def propagate_deadline(self, deadline):
        """Add a `deadline` attribute to the quote and propagate it to
        the quote's children by taking into account the duration of operations

        For instance if "self" has a duration of 5 and receives a deadline
        of 8, the quotes that "self" depends on will receive a deadline of
        8-5=3.
        """
        self.deadline = deadline
        children_deadline = deadline - self.step_duration
        if self.assembly_plan is not None:
            for segment, child in self.assembly_plan.items():
                child.propagate_deadline(children_deadline)


from collections import defaultdict
import textwrap
from ...config import SETTINGS
from .plot_tree_graph import plot_tree_graph
import matplotlib.font_manager as fm


def plot_supply_graph(
    quote=None,
    main_source=None,
    edges=None,
    levels=None,
    ax=None,
    textprops=None,
    margin=None,
    scale=1.0,
    interlevel_shift=0,
):
    """Plot the supply graph of all sources related to the provided source.
    Examples:
    ---------
    >>> quote.sources = chunks_assembly_station.compute_dict_supply_graph()
    Parameters
    ----------
    quote.sources
      A dictionary {"source_1_name": {"depth": 1, "infos": {...}},
                    "source_2_name": {"depth": 2, "infos": {...}},
                    etc.}
    ax
      A matplotlib ax. If None provided, one is created (and returned at the
      end)
    textprops
      Properties of the text. If none, a nice opensans font is used.
    
    Returns
    -------
    elements_positions, ax
      Dictionary of elements positions, matplotlib ax.
    """
    if main_source is not None:
        edges, levels = main_source.compute_supply_graph()
    if edges is None:
        edges = []
        levels = defaultdict(lambda *a: [])
        for source_name, infos in quote.sources.items():
            levels[infos._depth].append(source_name)
            for provider_name in infos.providers:
                edges.append((provider_name, source_name))
        levels = [levels[i] for i in range(max(levels) + 1)][::-1]
    else:
        sources_dict = {
            source.name: source
            for level in levels
            for source in level
        }

    fontawesome = fm.FontProperties(
        fname=SETTINGS["fontawesome-ttf-path"],
        size=18 * scale,
        family="sans-serif",
    )
    if textprops is None:
        textprops = fm.FontProperties(
            fname=SETTINGS["OpenSans-ttf-path"],
            size=12 * scale,
            family="sans-serif",
        )

    def draw_node(x, y, source_name, ax):
        if quote:
            source = quote.sources[source_name]
            symbol = source._report_fa_symbol
        else:
            source = sources_dict[source_name.name]
            symbol = source.report_fa_symbol
        
        ax.text(
            x,
            y,
            symbol,
            horizontalalignment="center",
            verticalalignment="center",
            fontproperties=fontawesome,
        )

        text = "\n".join(textwrap.wrap(source.name, 13))

        ax.text(
            x,
            y + 0.035 * scale ** 0.3,
            text,
            horizontalalignment="center",
            verticalalignment="bottom",
            fontproperties=textprops,
        )

    return plot_tree_graph(
        levels,
        edges,
        draw_node,
        ax=ax,
        scale=scale,
        interlevel_shift=interlevel_shift,
        margin=margin,
    )


from .DnaAssemblyMethod import DnaAssemblyMethod


class BluntEndAssemblyMethod(DnaAssemblyMethod):
    def compute_fragment_for_sequence_segment(self, sequence, segment):
        start, end = segment
        return sequence[start:end]



import numpy as np
from functools import lru_cache
from .SegmentSelector import SegmentSelector

try:
    import primer3

    PRIMER3_AVAILABLE = True
except ImportError:
    PRIMER3_AVAILABLE = False

class TmSegmentSelector(SegmentSelector):
    """Selects segment with melting temperature constraints.

    Given a location in a sequence, this selector will return a segment
    (=subsegment of the sequence) around this position, with size and
    temperature constraints.

    Can be used for Gibson Assembly, Golden Gate Assembly, etc.

    This class also implements methods to quickly compute all segments in a
    sequence. The result is then cached for speed.

    Parameters
    ----------
    min_size
      Minimal length of the segment, in nucleotides

    max_size
      Maximal length of the segment, in nucleotides

    min_tm
      Minimal melting temp allowed for the segments

    max_tm
      Maximal melting temp allowed for the segments

    precompute_segments
      If True, the whole sequence will be analyzed to find the best segment
      for each cutting point (and find cutting points which do not allow
      segments), using the AT/GC=2/4C heuristic for melting temperatures.

    primer3_params
      If ``precompute_segments`` is False and ``Primer3`` is installed, the
      melting temperature is computed
    """

    def __init__(
        self,
        min_size=18,
        max_size=22,
        min_tm=55,
        max_tm=65,
        precompute_segments=True,
        left_addition="",
        right_addition="",
        **primer3_params
    ):
        """Initialize."""
        self.min_size = min_size
        self.max_size = max_size
        self.precompute_segments = precompute_segments
        self.primer3_params = primer3_params
        self.min_tm = min_tm
        self.max_tm = max_tm
        self.middle_tm = 0.5 * (self.min_tm + self.max_tm)
        self.middle_size = int(0.5 * (self.min_size + self.max_size))
        self.left_addition = left_addition
        self.right_addition = right_addition
    
    @property
    def max_homology_size(self):
        return self.max_size

    def filter_location(self, sequence, index):
        """Return whether the sequence has a valid segment at this index."""
        if index == len(sequence):
            index = index - 1
        if self.precompute_segments:
            return self.compute_all_segments(sequence)[index] is not None
        else:
            return self.compute_optimal_segment(sequence, index)[1]

    def compute_segment_location(self, sequence, index):
        """Return the location (start, stop) of the selected segment."""
        if (index < 0) or (sequence == ''):
            return None
        if index == len(sequence):
            index = index - 1
        if self.precompute_segments:
            # calling a result that should be computed once then cached
            return self.compute_all_segments(sequence)[index]
        else:
            return self.compute_optimal_segment(sequence, index)[0]

    def compute_optimal_segment(self, sequence, index, init_size=None):
        """Return the best segment around some location.

        This function will be shortcut if ``precompute_segments`` is True,
        as it is slow to compute the best overhang for each location
        separately.
        """

        if init_size is None:
            init_size = self.middle_size

        def f(ovh_size):
            start, end = self.get_segment_coordinates(
                index, ovh_size, len(sequence)
            )
            tm = self.compute_tm(sequence[start:end])
            return (start, end), tm

        best_segment, best_tm = f(init_size)
        if best_tm < self.middle_tm:
            ovh_sizes = range(init_size, self.max_size)
        else:
            ovh_sizes = range(init_size, self.min_size, -1)
        for ovh_size in ovh_sizes:
            new_segment, new_tm = f(ovh_size)
            if abs(new_tm - self.middle_tm) < abs(best_tm - self.middle_tm):
                best_segment, best_tm = new_segment, new_tm
            else:
                break
        return best_segment, (self.min_tm < best_tm < self.max_tm)

    @lru_cache(maxsize=3)
    def compute_all_segments(self, sequence):
        """Quickly compute all segments in the sequence.

        This function uses the heuristic {A, T}=2degC, {G, C}=4degC to compute
        melting temperatures.

        This function uses vectorial operations for speed. The results are also
        cached.
        """

        if self.primer3_params == {}:
            lmin, lmax = self.min_size, self.max_size
            table = np.zeros((lmax + 1 - lmin, len(sequence)))

            # super fast GC computing using the trick 67=C, 71=G in unicode.
            seq_array = np.fromstring(sequence + "", dtype="uint8")
            cumsum = np.cumsum(2 + 2 * ((seq_array == 71) | (seq_array == 67)))
            for i, oh_size in enumerate(range(lmin, min(len(sequence) - 1, lmax + 1))):
                arr = cumsum[oh_size:] - cumsum[:-oh_size]
                start = int(oh_size / 2)
                end = start + len(arr)
                table[i, start:end] = arr
                # print (i, start, table)
                table[i, :start] = table[i, start]
                table[i, end:] = table[i, end - 1]
            scores = -(table - self.min_tm) * (table - self.max_tm)
            best_sizes_indices = scores.argmax(axis=0)
            best_sizes = lmin + best_sizes_indices
            validities = np.choose(best_sizes_indices, scores) >= 0
            osizes_and_validities = zip(best_sizes, validities)
        else:
            osizes_and_validity = [self.compute_optimal_segment(sequence, 0)]
            for i in range(1, len(sequence)):
                init_size = osizes_and_validity[-1][0]
                ovh, val = self.find_optimal_segment(sequence, i, init_size)
                osizes_and_validity.append((len(ovh), val))
        return [
            None
            if not valid
            else self.get_segment_coordinates(i, ovh_size, len(sequence))
            for i, (ovh_size, valid) in enumerate(osizes_and_validities)
        ]

    def compute_tm(self, sequence):
        """Return the melting temp of the sequence.

        If Primer3 is available, it's internal melting temperature calculator
        is used with ``self.primer3_params`` used as parameters.
        Else the heuristic AT/GC=2/4C is used.
        """
        if self.primer3_params == {}:
            return sum([4 if c in "GC" else 2 for c in sequence])
        if not PRIMER3_AVAILABLE:
            raise ImportError(
                "Melting temperature computation with '%s' "
                "Requires Primer3 installed."
                % self.primer3_params.get("method", "[unknown method]")
            )
        return primer3.calcTm(sequence, **self.primer3_params)
    

    def __str__(self):
        result = "Tm(%d-%dC, %d-%dbp)" % (
            self.min_tm, self.max_tm, self.min_size, self.max_size)
        if self.left_addition:
            result = ('...%s-' % self.left_addition[-12:]) + result
        if self.right_addition:
            result = result + ('-%s...' % self.right_addition[:12]) 
        return result


import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from .matplotlib_export import matplotlib_figure_to_svg_base64_data

def give_quotes_html_locations(quotes, len_sequence, ax=None):
    """Quickly adds an html-location field containing an HTML image of the
    location of each sub-quote in the final sequence.

    The image is a line representing the full sequence and a rectangle
    indicating the position of the subquote in the the full sequence.

    The advantage of processing all quotes at once is that the
    figure is drawn once, then only the position of the rectangle and the
    title are modified.
    """
    if ax is None:
        _fig, ax = plt.subplots(1, figsize=(1.5, 0.2), facecolor=None)
    ax.set_frame_on(False)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    ax.plot([0, len_sequence], [0, 0], c="k")
    rect = mpatches.Rectangle(
        (0, -0.5),
        0,
        1,
        color="#ddddff",
        ec="k",
        alpha=1,
        zorder=1000,
        transform=ax.transData,
    )
    ax.add_patch(rect)
    ax.set_ylim((-1, 1))
    ax.set_xlim(0, len_sequence)
    text = ax.text(0.5, 0.65, "none", horizontalalignment="center", fontsize=4)
    for q in quotes:
        location = q.final_location
        if location is None:
            q.html_location = ""
        else:
            rect.set_bounds(location[0], -0.5, location[1] - location[0], 1.0)
            text.set_text("%d-%d" % (location[0], location[1]))
            svg_data = matplotlib_figure_to_svg_base64_data(
                ax.figure, transparent=True
            )
            q.html_location = "<img src='%s' />" % svg_data
    plt.close(ax.figure)


from copy import deepcopy
from .ObjectDict import ObjectDict
from . import mixins


class AssemblyPlanReport(
    mixins.PlotsMixin,
    mixins.FolderReportMixin,
    mixins.GenbankExportMixin,
    mixins.PdfReportMixin,
):
    def __init__(self, plan, sources):
        self.plan = ObjectDict.from_dict(plan)
        self.sources = ObjectDict.from_dict(sources)

    @staticmethod
    def from_dnaweaver_quote(quote):
        plan = quote.assembly_plan_as_dict()
        sources = quote.source.dict_supply_graph()
        return AssemblyPlan(plan, sources)

    def to_steps_list(self):
        plan = deepcopy(self.plan)
        nodes = []

        def rec(node, depth=0):
            if node.get("_visited", False):
                return
            node["_visited"] = True
            assembly_plan = node.get("assembly_plan", [])
            node["children"] = [n["id"] for n in assembly_plan]
            nodes.append(node)
            for other in sorted(
                assembly_plan, key=lambda n: n["segment_start"]
            ):
                rec(other)

        rec(plan)
        return nodes



# This will try to import setuptools. If not here, it will reach for the
# embedded ez_setup (or the ez_setup package). If none, it fails with a message
try:
    from setuptools import setup
except ImportError:
    try:
        import ez_setup

        ez_setup.use_setuptools()
    except ImportError:
        raise ImportError(
            "DnaWeaver could not be installed, probably because"
            " neither setuptools nor ez_setup are installed on"
            "this computer. \nInstall ez_setup "
            "([sudo] pip install ez_setup) and try again."
        )

from setuptools import setup, find_packages

exec(open("dnaweaver/version.py").read())  # loads __version__

setup(
    name="DnaWeaver",
    version=__version__,
    author="Zulko",
    description="Make ordering and assembly plans for DNA sequences",
    url='https://github.com/Edinburgh-Genome-Foundry/DnaWeaver',
    long_description=open("pypi-readme.rst").read(),
    license="MIT",
    keywords="DNA optimization assembly ordering synthetic biology",
    packages=find_packages(exclude="docs"),
    include_package_data=True,
    install_requires=[
        "numpy",
        "Biopython",
        "proglog",
        "networkx",
        "flametree",
        "dna_features_viewer",
        "weasyprint",
        "pandas",
        "jinja2"
    ],
)



from ...DnaQuote import DnaQuote
from ..DnaSupplier import DnaSupplier


class DnaSuppliersComparator(DnaSupplier):
    """Special source that compares quotes from other DnaSuppliers.

    Upon receiving a sequence, that source will submit the sequence to
    Downstream sources, which deliver each one optimal quote. The comparator
    then returns the one quote with the lowest price.

    Parameters
    ----------

    suppliers
      List of `DnaSuppliers` that

    memoize
      Whether the quotes should be kept in memory to avoid re-computing the
      same quote several times. Can accelerate computations but is
      RAM-expensive.
    """

    class_description = "DNA sources comparator"
    operation_type = "comparison"
    report_fa_symbol = u""
    report_fa_symbol_plain = u"circle-o"
    report_color = "#000000"

    def __init__(
        self,
        suppliers=(),
        memoize=False,
        sequence_constraints=(),
        return_first_accepted_quote=False,
        name="comparator",
    ):
        self.set_suppliers(suppliers)
        self.memoize = memoize
        self.sequence_constraints = sequence_constraints
        self.return_first_accepted_quote = return_first_accepted_quote
        self.memoize_dict = {}
        self.name = name

    def get_best_price(
        self,
        sequence,
        max_lead_time=None,
        with_assembly_plan=False,
    ):
        """Returns a price-optimal DnaQuote for the given sequence.

        Parameters
        ----------

        sequence (str)
          The sequence submitted to the Dna Source for a quots

        max_lead_time (float)
          If provided, the quote returned is the best quote (price-wise) whose
          lead time is less or equal to max_lead_time.

        with_assembly_plan
          If True, the assembly plan is added to the quote
        """

        best_quote = None
        best_basepair_price = 1e8
        for source in sorted(
            self.suppliers, key=lambda s: s.min_basepair_price
        ):
            # print (self.name, len(sequence), source)
            if source.min_basepair_price > best_basepair_price:
                break
            quote = source.get_quote(
                sequence,
                max_lead_time=max_lead_time,
                with_assembly_plan=with_assembly_plan
            )
            # print ('=>', quote.accepted, quote.price)
            if not quote.accepted:
                continue
            if self.return_first_accepted_quote:
                best_quote = quote
                break
            quote_basepair_price = quote.price / float(len(sequence))
            if quote_basepair_price < best_basepair_price:
                best_quote = quote
                best_basepair_price = quote_basepair_price
        if best_quote is None:
            return DnaQuote(
                self,
                sequence,
                accepted=False,
                message="Sequence was rejected by all sources.",
            )
        return best_quote

    def additional_dict_description(self):
        return {
            "dna sources": [source.name for source in self.suppliers],
        }

    @staticmethod
    def from_dict(data):
        return DnaSuppliersComparator(
            name=data["name"], suppliers=data["suppliers"]
        )

    def set_suppliers(self, suppliers):
        self.suppliers = suppliers
        if len(suppliers):
            self.min_basepair_price = min(
                [supplier.min_basepair_price for supplier in suppliers]
            )
        else:
            self.min_basepair_price = None

    def suggest_cuts(self, sequence):
        return sorted(
            set(
                [
                    cut
                    for supplier in self.suppliers
                    if hasattr(supplier, "suggest_cuts")
                    for cut in supplier.suggest_cuts(sequence)
                ]
            )
        )
